/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

class CompetitionController {

  def competitionControlService
  def jmsBrokerManagementService
  static allowedMethods = [save: "POST", update: "POST", delete: "POST"]

  def index = {
    redirect(action: "list", params: params)
  }

  def list = {
    params.max = Math.min(params.max ? params.int('max') : 10, 100)
    [competitionInstanceList: Competition.list(params), competitionInstanceTotal: Competition.count()]
  }

  def create = {
    def competitionInstance = new Competition()
    competitionInstance.properties = params
    return [competitionInstance: competitionInstance]
  }

  def save = {
    if (Competition.count() > 0) {
      flash.error = "Currently TAC Energy can only run a single competition. Please re-use the existing competition instead of creating a new one."
      redirect(action: "list")
    } else {
      def competitionInstance = new Competition(params)
      if (competitionInstance.save(flush: true)) {
        competitionInstance = competitionControlService.initCompetition(competitionInstance)
        flash.message = "${message(code: 'default.created.message', args: [message(code: 'competition.label', default: 'Competition'), competitionInstance.id])}"
        redirect(action: "show", id: competitionInstance.id)
      }
      else {
        render(view: "create", model: [competitionInstance: competitionInstance])
      }
    }
  }

  def start = {
    def competitionInstance = Competition.get(params.id)
    if (!competitionInstance) {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    } else if (Competition.findByCompetitionStatus(CompetitionStatus.Running)) {
      flash.error = "${message(code: 'competition.already.running.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "show", id: params.id)
    }
    else {
      try {
        competitionControlService.start(competitionInstance)
        flash.message = "Competition '${competitionInstance}' started"
        redirect(action: "show", id: params.id)
      } catch (Exception e) {
        flash.error = e.message
        redirect(action: "show", id: params.id)
      }
    }
  }

  def stats = {
    [stats: jmsBrokerManagementService.getStats()]
  }

  def stop = {
    def competitionInstance = Competition.get(params.id)
    if (!competitionInstance) {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    }
    else if (competitionInstance.competitionStatus != CompetitionStatus.Running) {
      flash.error = "${message(code: 'competition.not.running.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "show", id: params.id)
    }
    else {
      try {
        competitionControlService.stop(competitionInstance)
        flash.message = "Competition '${competitionInstance}' stopped"
        redirect(action: "show", id: params.id)
      } catch (Exception e) {
        flash.error = e.message
        redirect(action: "show", id: params.id)
      }
    }
  }

  def reset = {
    def competitionInstance = Competition.get(params.id)
    if (!competitionInstance) {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    }
    else {
      try {
        competitionControlService.reset(competitionInstance)
        flash.message = "Competition '${competitionInstance}' reset"
        redirect(action: "show", id: params.id)
      } catch (CompetitionManagementException ex) {
        flash.error = "Please disconnect all clients before resetting the competition."
        redirect(action: "show", id: params.id)
      } catch (Exception e) {
        flash.error = e.message
        redirect(action: "show", id: params.id)
      }
    }
  }

  def show = {
    def competitionInstance = Competition.get(params.id)
    if (!competitionInstance) {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    }
    else {
      [competitionInstance: competitionInstance]
    }
  }

  def edit = {
    def competitionInstance = Competition.get(params.id)
    if (!competitionInstance) {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    }
    else {
      return [competitionInstance: competitionInstance]
    }
  }

  def update = {
    def competitionInstance = Competition.get(params.id)
    if (competitionInstance) {
      if (params.version) {
        def version = params.version.toLong()
        if (competitionInstance.version > version) {

          competitionInstance.errors.rejectValue("version", "default.optimistic.locking.failure", [message(code: 'competition.label', default: 'Competition')] as Object[], "Another user has updated this Competition while you were editing")
          render(view: "edit", model: [competitionInstance: competitionInstance])
          return
        }
      }
      competitionInstance.properties = params
      if (!competitionInstance.hasErrors() && competitionInstance.save(flush: true)) {
        competitionControlService.reset(competitionInstance)
        flash.message = "${message(code: 'competition.updatereset.message', args: [message(code: 'competition.label', default: 'Competition.'), competitionInstance.id])}"
        redirect(action: "show", id: competitionInstance.id)
      }
      else {
        render(view: "edit", model: [competitionInstance: competitionInstance])
      }
    }
    else {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    }
  }

  def delete = {
    def competitionInstance = Competition.get(params.id)
    if (competitionInstance) {
      try {
        competitionControlService.delete(competitionInstance)
        flash.message = "${message(code: 'default.deleted.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
        redirect(action: "list")
      }
      catch (Exception e) {
        flash.error = "${message(code: 'default.not.deleted.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
        log.error "Competition cannot be deleted: ${e.message} (${e.cause})"
        redirect(action: "show", id: params.id)
      }
    }
    else {
      flash.error = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    }
  }
}
