/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

class PersonController {

  def authenticateService
  def idGeneratorService

  // the delete, save and update actions only accept POST requests
  static Map allowedMethods = [delete: 'POST', save: 'POST', update: 'POST']

  def index = {
    redirect action: list, params: params
  }

  def list = {
    if (!params.max) {
      params.max = 10
    }
    [personList: Person.list(params)]
  }

  def chart = {
    def competitionInstance = idGeneratorService.getCompetition()
    if (!competitionInstance) {
      flash.message = "${message(code: 'default.not.found.message', args: [message(code: 'competition.label', default: 'Competition'), params.id])}"
      redirect(action: "list")
    } else {
      def personInstance = Person.get(params.id)
      if (!personInstance) {
        flash.message = "${message(code: 'default.not.found.message', args: [message(code: 'person.label', default: 'Person'), params.id])}"
        redirect(action: "list")
      } else {
        def products = Product.findAllByCompetition(competitionInstance, [sort: 'serialNumber', order: 'asc'])
        if (!products) {
          flash.message "Cannot create chart because no products were found."
          redirect (action: 'show', id: params.id)
        } else {
          List forecastList = Forecast.withCriteria {
            eq('person',personInstance)
            eq('competition', competitionInstance)
            eq('competitionTime', competitionInstance.currentCompetitionTime)
          }
          def depotPositionList = DepotPosition.withCriteria {
            eq('competition', competitionInstance)
            eq('person', personInstance)
            eq('latest', true)
          }
          def forecastDemand = []
          def realDemand = []
          def acquiredDemand = []
          products?.each {product ->
            def forecastInstance = forecastList.find {it.product == product}
            if (forecastInstance) {
              forecastDemand.add([product.startDateTime.getTime(), forecastInstance.forecastValue])
              realDemand.add([product.startDateTime.getTime(), forecastInstance.trueValue])
              forecastList.remove(forecastInstance)
            }
            def depotPosition = depotPositionList.find {it.product = product}
            if (depotPosition) {
              acquiredDemand.add([product.startDateTime.getTime(), depotPosition.balance])
              depotPositionList.remove(depotPosition)
            } 
          }
          [competitionInstance: competitionInstance,
           personInstance: personInstance,
           forecastDemand: forecastDemand,
           reaLDemand: realDemand,
           acquiredDemand: acquiredDemand
          ]
        }
      }
    }
  }

  def show = {
    def person = Person.get(params.id)
    if (!person) {
      flash.message = "Person not found with id $params.id"
      redirect action: list
      return
    }
    List roleNames = []
    for (role in person.authorities) {
      roleNames << role.authority
    }
    roleNames.sort { n1, n2 ->
      n1 <=> n2
    }
    [person: person, roleNames: roleNames]
  }

  /**
   * Person delete action. Before removing an existing person,
   * he should be removed from those authorities which he is involved.
   */
  def delete = {

    def person = Person.get(params.id)
    if (person) {
      def authPrincipal = authenticateService.principal()
      //avoid self-delete if the logged-in user is an admin
      if (!(authPrincipal instanceof String) && authPrincipal.username == person.username) {
        flash.message = "You can not delete yourself, please login as another admin and try again"
      }
      else {
        //first, delete this person from People_Authorities table.
        Role.findAll().each { it.removeFromPeople(person) }
        person.delete()
        flash.message = "Person $params.id deleted."
      }
    }
    else {
      flash.message = "Person not found with id $params.id"
    }

    redirect action: list
  }

  def edit = {

    def person = Person.get(params.id)
    if (!person) {
      flash.message = "Person not found with id $params.id"
      redirect action: list
      return
    }

    return buildPersonModel(person)
  }

  /**
   * Person update action.
   */
  def update = {

    def person = Person.get(params.id)
    if (!person) {
      flash.message = "Person not found with id $params.id"
      redirect action: edit, id: params.id
      return
    }

    long version = params.version.toLong()
    if (person.version > version) {
      person.errors.rejectValue 'version', "person.optimistic.locking.failure",
          "Another user has updated this Person while you were editing."
      render view: 'edit', model: buildPersonModel(person)
      return
    }

    def oldPassword = person.passwd
    person.properties = params
    if (!params.passwd.equals(oldPassword)) {
      person.passwd = authenticateService.encodePassword(params.passwd)
    }
    if (person.save()) {
      Role.findAll().each { it.removeFromPeople(person) }
      addRoles(person)
      redirect action: show, id: person.id
    }
    else {
      render view: 'edit', model: buildPersonModel(person)
    }
  }

  def create = {
    [person: new Person(params), authorityList: Role.list()]
  }

  /**
   * Person save action.
   */
  def save = {

    def person = new Person()
    person.properties = params
    person.passwd = authenticateService.encodePassword(params.passwd)
    if (person.save()) {
      addRoles(person)
      redirect action: show, id: person.id
    }
    else {
      render view: 'create', model: [authorityList: Role.list(), person: person]
    }
  }

  private void addRoles(person) {
    for (String key in params.keySet()) {
      if (key.contains('ROLE') && 'on' == params.get(key)) {
        Role.findByAuthority(key).addToPeople(person)
      }
    }
  }

  private Map buildPersonModel(person) {

    List roles = Role.list()
    roles.sort { r1, r2 ->
      r1.authority <=> r2.authority
    }
    Set userRoleNames = []
    for (role in person.authorities) {
      userRoleNames << role.authority
    }
    LinkedHashMap<Role, Boolean> roleMap = [:]
    for (role in roles) {
      roleMap[(role)] = userRoleNames.contains(role.authority)
    }

    return [person: person, roleMap: roleMap]
  }
}