/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

class AuctionService {

  boolean transactional = true

  def idGeneratorService
  def queueOutputService

  public Shout addShout(Shout shout) throws OrderEntryException, MarketOrderException, ExecutionException, SettlementException {
    /* set shoutID
    * set transactionID
    * set ModReasonCode = INSERT
    * validate()
    * save()
    * call matching
    */

    shout.setShoutID(idGeneratorService.nextId())
    shout.setTransactionID(shout.getShoutID())
    shout.setModReasonCode(ModReasonCode.INSERT)
    if (!shout.save()) throw new OrderEntryException("Failed to add shout: ${shout.errors}")
    return matching(shout)
  }

  public Shout deleteShout(Shout shout) throws OrderEntryException, MarketOrderException, ExecutionException, SettlementException {

    /*
    * set ModReasonCode = DELETIONBYUSER
    * validate()
    * call deletion
    */
    if (shout.outdated) throw new OrderUpdateException("Cannot delete outdated shout: ${shout}")
    shout.setOutdated(true)
    if (!shout.save()) throw new OrderEntryException("Failed to delete shout: ${shout.errors}")

    def delShout = shout.initModification(ModReasonCode.DELETIONBYUSER)
    delShout.setTransactionID(idGeneratorService.nextId())
    delShout.save()
    updateOrderbook(delShout)
    return delShout
  }

  public Shout editShout(Shout shout) throws OrderEntryException, OrderUpdateException, MarketOrderException, ExecutionException, SettlementException {
    /*
    * set ModReasonCode = MODIFICATION
    * validate()
    * call edit
    */

    if (shout.outdated) throw new OrderUpdateException("Cannot edit outdated shout: ${shout}")

    def oldShout = Shout.get(shout.id)
    deleteShout(oldShout)

    def newShout = shout.initModification(ModReasonCode.MODIFICATION)
    newShout.setTransactionID(idGeneratorService.nextId())
    if (!newShout.save()) throw new OrderEntryException("Failed to edit shout: ${shout.errors}")
    newShout = matching(newShout)
    return newShout
  }

  private Shout matching(Shout shout) throws MarketOrderException, ExecutionException, SettlementException {
    /*
    * loop: get candidate list for matching
    * call execute(shout, candidateShout)
    * result: shout(with remainingQuantity)
    * call orderbookUpdate()
    *
    */

    //get matching candidates in correct order
    def shoutCandidates = Shout.withCriteria {
      eq('product', shout.getProduct())
      eq('outdated', false)
      if (shout.getBuySellIndicator() == BuySellIndicator.BUY) {
        if (shout.getOrderType() == OrderType.LIMIT) {
          le('limitPrice', shout.getLimitPrice())
        }
        eq('buySellIndicator', BuySellIndicator.SELL)
        and {
          order("limitPrice", "asc")
          order("dateMod", "asc")
        }

      }
      else if (shout.getBuySellIndicator() == BuySellIndicator.SELL) {
        if (shout.getOrderType() == OrderType.LIMIT) {
          ge('limitPrice', shout.getLimitPrice())
        }
        eq('buySellIndicator', BuySellIndicator.BUY)
        and {
          order("limitPrice", "desc")
          order("dateMod", "asc")
        }

      }
      order("shoutID", "asc")
    }

    //loop until initial shout is fully executed or every shoutCandidate matched
    for (Shout shoutCandidate in shoutCandidates) {
      shout = execute(shout, shoutCandidate)
      if (shout.getModReasonCode() == ModReasonCode.EXECUTION) {
        break;
      }
    }

    //delete market order if it cannot be executed completely
    if (shout.getModReasonCode() != ModReasonCode.EXECUTION && shout.getOrderType() == OrderType.MARKET) {
      shout.outdated = true
      if (!shout.save(flush:true)) throw new MarketOrderException("Failed to set MarketOrder outdated")
      def delShout = shout.initModification(ModReasonCode.DELETIONBYSYSTEM)
      delShout.transactionID = shout.getTransactionID()
      if (!delShout.save()) throw new MarketOrderException("Failed to set (remaining) MarketOrder to deleted after it was matched only partially.")
      shout = delShout
    }

    //call orderbookUpdate
    updateOrderbook(shout)
    return shout
  }

  private Shout execute(Shout shout, Shout candidate) throws ExecutionException, SettlementException {
    /* partialExecution
   * or completeExecution
   *
   * write TransactionLog
    */
    def execQuantity = 0.0
    def modShout
    def modCandidate

    //decide execQuantity (=lower remainingQuantity)
    execQuantity = shout.getRemainingQuantity() <= candidate.getRemainingQuantity() ? shout.getRemainingQuantity() : candidate.getRemainingQuantity()

    //clone initialShout and update properties
    modShout = shout.initModification(execQuantity == shout.getRemainingQuantity() ? ModReasonCode.EXECUTION : ModReasonCode.PARTIALEXECUTION)
    modShout.setRemainingQuantity(modShout.getRemainingQuantity() - execQuantity)
    modShout.setExecutionPrice(candidate.getLimitPrice())
    modShout.setExecutionQuantity(execQuantity)
    modShout.setOutdated(execQuantity == shout.getRemainingQuantity())
    modShout.setTransactionID(shout.getTransactionID())
    if (!modShout.save()) throw new ExecutionException("Failed to execute incoming order")

    //clone candidateShout and update properties
    modCandidate = candidate.initModification(execQuantity == candidate.getRemainingQuantity() ? ModReasonCode.EXECUTION : ModReasonCode.PARTIALEXECUTION)
    modCandidate.setRemainingQuantity(modCandidate.getRemainingQuantity() - execQuantity)
    modCandidate.setExecutionPrice(candidate.getLimitPrice())
    modCandidate.setExecutionQuantity(execQuantity)
    modCandidate.setOutdated(execQuantity == candidate.getRemainingQuantity())
    modCandidate.setTransactionID(shout.getTransactionID())
    if (!modCandidate.save()) throw new ExecutionException("Failed to execute order in orderbook")
    
    //set initial shout and candidate shout outdated + save
    shout.setOutdated(true)
    if (!shout.save()) throw new ExecutionException("Failed to set old incoming order outdated after execution")
    queueOutputService.sendOrderstatus(shout)
    queueOutputService.sendOrderstatus(modShout)
    candidate.setOutdated(true)
    if (!candidate.save()) throw new ExecutionException("Failed to set old order in oderbook outdated after execution")
    queueOutputService.sendOrderstatus(candidate)
    queueOutputService.sendOrderstatus(modCandidate)

    //write TransactionLog: Trade
    def execTransactionLog = new TransactionLog(
        product: shout.getProduct(),
        transactionID: shout.getTransactionID(),
        type: TransactionType.TRADE,
        price: candidate.getLimitPrice(),
        quantity: execQuantity,
        buyerID: shout.getBuySellIndicator() == BuySellIndicator.BUY ? shout.getPerson() : candidate.getPerson(),
        sellerID: shout.getBuySellIndicator() == BuySellIndicator.SELL ? shout.getPerson() : candidate.getPerson(),
        buySellIndicator: shout.getBuySellIndicator())

    if (!execTransactionLog.save()) throw new ExecutionException("Failed to create Trade TransactionLog")
    queueOutputService.sendTradeNotification(execTransactionLog)
    executeSettlement(execTransactionLog)

    return modShout
  }


  private void executeSettlement(TransactionLog tl) throws SettlementException {
    def seller = tl.sellerID
    def buyer = tl.buyerID
    def cash = (tl.quantity * tl.price)
    def oldSellerCashBalance
    def oldBuyerCashBalance
    def oldSellerDepotBalance
    def oldBuyerDepotBalance

    def oldSellerCashPosition = CashPosition.findByPersonAndLatest(seller, true)
    if (oldSellerCashPosition) {
      oldSellerCashBalance = oldSellerCashPosition.getBalance()
      oldSellerCashPosition.setLatest(false)
      if (!oldSellerCashPosition.save()) throw new SettlementException("Failed to set old SellerCashPosition outdated")
    } else {
      oldSellerCashBalance = 0
    }

    def cashPositionSeller = new CashPosition(person: seller, change: cash, transactionID: tl.transactionID, balance: (oldSellerCashBalance + cash))
    if (!cashPositionSeller.save()) throw new SettlementException("Failed to create CashPosition for Seller")
    queueOutputService.sendCash(cashPositionSeller)

    def oldBuyerCashPosition = CashPosition.findByPersonAndLatest(buyer, true)
    if (oldBuyerCashPosition) {
      oldBuyerCashBalance = oldBuyerCashPosition.getBalance()
      oldBuyerCashPosition.setLatest(false)
      if (!oldBuyerCashPosition.save()) throw new SettlementException("Failed to set old BuyerCashPosition outdated")
    } else {
      oldBuyerCashBalance = 0
    }

    def cashPositionBuyer = new CashPosition(person: buyer, change: -cash, transactionID: tl.transactionID, balance: (oldBuyerCashBalance - cash))
    if (!cashPositionBuyer.save()) throw new SettlementException("Failed to create CashPosition for Buyer")
    queueOutputService.sendCash(cashPositionBuyer)


    def oldSellerDepotPosition = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', seller)
      eq('product', tl.product)
      eq('latest', true)
    }

    if (oldSellerDepotPosition) {
      oldSellerDepotBalance = oldSellerDepotPosition.getBalance()
      oldSellerDepotPosition.setLatest(false)
      if (!oldSellerDepotPosition.save()) throw new SettlementException("Failed to set old SellerDepotPosition outdated")
    } else {
      oldSellerDepotBalance = 0.0
    }

    def depotPositionSeller = new DepotPosition(person: seller, product: tl.product, change: -tl.quantity, transactionID: tl.transactionID, balance: (oldSellerDepotBalance - tl.quantity))
    if (!depotPositionSeller.save()) throw new SettlementException("Failed to create DepotPosition for Seller")
    queueOutputService.sendDepot(depotPositionSeller)


    def oldBuyerDepotPosition = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', buyer)
      eq('product', tl.product)
      eq('latest', true)
    }

    if (oldBuyerDepotPosition) {
      oldBuyerDepotBalance = oldBuyerDepotPosition.getBalance()
      oldBuyerDepotPosition.setLatest(false)
      if (!oldBuyerDepotPosition.save()) throw new SettlementException("Failed to set old BuyerDepotPosition outdated")
    } else {
      oldBuyerDepotBalance = 0.0
    }

    def depotPositionBuyer = new DepotPosition(person: buyer, product: tl.product, change: tl.quantity, transactionID: tl.transactionID, balance: (oldBuyerDepotBalance + tl.quantity))

    if (!depotPositionBuyer.save()) throw new SettlementException("Failed to create DepotPosition Buyer")
    queueOutputService.sendDepot(depotPositionBuyer)
  }

  private void updateOrderbook(Shout shout) throws OrderbookException {
    /*
    * get top ten from db
    * updateOrderbook
    *
    * if change in quote write TransactionLog
    */
    boolean firstOrderbook

    Orderbook latestOrderbook = Orderbook.withCriteria(uniqueResult: true) {
      maxResults(1)
      eq('product', shout.getProduct())
      eq('outdated', false)
    }
    if (latestOrderbook == null) {
      latestOrderbook = new Orderbook()
      firstOrderbook = true
    }

    def bestBids = Shout.executeQuery("SELECT limitPrice, SUM(remainingQuantity) FROM Shout WHERE product = " + shout.getProduct().id + " AND buySellIndicator = 1 AND outdated = 'false' GROUP BY  limitPrice ORDER BY limitPrice DeSC LIMIT 1,10")
    def bestAsks = Shout.executeQuery("SELECT limitPrice, SUM(remainingQuantity) FROM Shout WHERE product = " + shout.getProduct().id + " AND buySellIndicator = -1 AND outdated = 'false' GROUP BY  limitPrice ORDER BY limitPrice ASC LIMIT 1,10")

    BigDecimal[][][] newOrderbookArray = new BigDecimal[2][2][10]
    BigDecimal[][][] latestOrderbookArray = latestOrderbook.getOrderbookArray()
    boolean orderbookChangeFound = false
    int levelCounter = 0

    //no open bid orders left
    if (bestBids.size() == 0) {
      //empty bid orderbook is new situation
      if (latestOrderbook.getBid1() != null) {
        orderbookChangeFound = true
      }
    }
    else {
      //BID: check if new orderbook entry && create newOrderbookArray
      levelCounter = 0

      while (levelCounter <= 9) {
        //create newOrderbookArray bid entry

        //price
        if (bestBids[levelCounter]) {
          newOrderbookArray[0][0][levelCounter] = bestBids[levelCounter][0] //price
          newOrderbookArray[0][1][levelCounter] = bestBids[levelCounter][1] //size
        } else {
          //fill empty levels (in case of deletion this is necessary)
          newOrderbookArray[0][0][levelCounter] = null
          newOrderbookArray[0][1][levelCounter] = 0.0
        }

        //changed?
        if (!orderbookChangeFound && !(latestOrderbookArray[0][0][levelCounter] == newOrderbookArray[0][0][levelCounter] && latestOrderbookArray[0][1][levelCounter] == newOrderbookArray[0][1][levelCounter])) {
          orderbookChangeFound = true
        }
        levelCounter++

      }
    }

    //no open ask orders left
    if (bestAsks.size() == 0) {
      //empty ask orderbook is new situation
      if (latestOrderbook.getAsk1() != null) {
        orderbookChangeFound = true
      }
    }
    else {
      //ASK: check if new orderbook entry && create newOrderbookArray
      levelCounter = 0

      while (levelCounter <= 9) {
        //create newOrderbookArray bid entry

        //price
        if (bestAsks[levelCounter]) {
          newOrderbookArray[1][0][levelCounter] = bestAsks[levelCounter][0] //price
          newOrderbookArray[1][1][levelCounter] = bestAsks[levelCounter][1] //size
        } else {
          //fill empty levels (in case of deletion this is necessary)
          newOrderbookArray[1][0][levelCounter] = null
          newOrderbookArray[1][1][levelCounter] = 0.0
        }

        //changed?
        if (!orderbookChangeFound && !(latestOrderbookArray[1][0][levelCounter] == newOrderbookArray[1][0][levelCounter] && latestOrderbookArray[1][1][levelCounter] == newOrderbookArray[1][1][levelCounter])) {
          orderbookChangeFound = true
        }
        levelCounter++

      }
    }

    //create new orderbook entry
    if (orderbookChangeFound) {
      def newOrderbook = new Orderbook(product: shout.getProduct(), transactionID: shout.getTransactionID(), dateExecuted: shout.getDateMod())
      newOrderbook.setOrderbookArray(newOrderbookArray)
      if (!newOrderbook.save()) throw new OrderbookException("Failed to create Orderbook")
      queueOutputService.sendOrderbookNotification(newOrderbook)

      //Set old orderbook entry outdated
      if (!firstOrderbook) {
        latestOrderbook.setOutdated(true)
        if (!latestOrderbook.save()) throw new OrderbookException("Failed to set old Orderbook outdated")
      }

      //check if new quote
      //get latest TransactionLog
      TransactionLog latestTransactionLog = TransactionLog.withCriteria(uniqueResult: true) {
        maxResults(1)
        eq('product', shout.getProduct())
        eq('type', TransactionType.QUOTE)
        order("id", "desc")
      }
      if (latestTransactionLog == null) {
        latestTransactionLog = new TransactionLog(bidSize: 0.0, askSize: 0.0)
      }

      //create new TransactionLog
      def newTransactionLog = new TransactionLog(product: shout.getProduct(), transactionID: shout.getTransactionID(), type: TransactionType.QUOTE)
      newTransactionLog.setBid(newOrderbookArray[0][0][0])
      newTransactionLog.setBidSize(newOrderbookArray[0][1][0] == null ? 0 : newOrderbookArray[0][1][0])
      newTransactionLog.setAsk(newOrderbookArray[1][0][0])
      newTransactionLog.setAskSize(newOrderbookArray[1][1][0] == null ? 0 : newOrderbookArray[1][1][0])

      if (latestTransactionLog.getBid() != newTransactionLog.getBid()
          || latestTransactionLog.getBidSize() != newTransactionLog.getBidSize()
          || latestTransactionLog.getAsk() != newTransactionLog.getAsk()
          || latestTransactionLog.getAskSize() != newTransactionLog.getAskSize()) {
        if (!newTransactionLog.save()) throw new OrderbookException("Failed to create Quote TransactionLog")
        queueOutputService.sendQuoteNotification(newTransactionLog)
      }
    }
  }


}
