/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

class CompetitionControlService {

  boolean transactional = true

  private static final Random rand = new Random(new GregorianCalendar().getTimeInMillis())

  def idGeneratorService
  def productManagementService
  def queueOutputService
  def queueManagementService
  def forecastGenerationService
  def jmsBrokerManagementService
  def quartzScheduler

  /* ---------------
   * Public methods
   * ---------------
   */

  public Competition initCompetition(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot init competition ${competition}")

    competition.current = true
    competition.competitionStatus = CompetitionStatus.Initialized
    competition.save(flush: true)
    productManagementService.initProducts(competition) //initialize products (Timeslots) for this competition
    createParticipants(competition)
    setQueueNames(competition)
    forecastGenerationService.initForecasts(competition)
    jmsBrokerManagementService.initTopicsAndQueues()
    return competition
  }

  public void start(Competition competition) throws PersonEnableException, CompetitionManagementException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot start competition ${competition}")
    try {
      log.debug('Starting competition...')
      enableParticipants(competition, true)
      log.debug('Participants enabled...')
      queueOutputService.sendCompetitionNotification(competition, ActionType.Started)
      Thread.currentThread().sleep(2000) //Sleep for one second in order to ensure that the competition notification is received by all participants first
      log.debug('Comepetition start announced...')
      productManagementService.openInitialProducts(competition)
      log.debug('Initial products opened')
      competition.enabled = true
      competition.competitionStatus = CompetitionStatus.Running
      competition.save()
      log.debug('Competition status updated...')
      reScheduleProductAdvanceJob(competition)
      log.debug('CronJob for product shifting scheduled...')
    } catch (Exception e) {
      log.warn "An Exception occurred while starting competition ${competition}: ${e.message}"
      throw new CompetitionManagementException("An Exception occurred while starting competition ${competition}",e)
    }
  }

  public void stop(Competition competition) throws CompetitionManagementException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot stop competition ${competition}")
    try {
      if (competition.competitionStatus == CompetitionStatus.Running) queueOutputService.sendCompetitionNotification(competition, ActionType.Stopped)
      log.debug('Competition stop announced...')
      quartzScheduler.pauseTrigger('mySimpleTrigger', 'mySimpleTriggerGroup')
      log.debug('Quartz scheduler stopped...')
      enableParticipants(competition, false)
      log.debug('Participants disabled...')
      competition.enabled = false
      competition.current = false
      competition.competitionStatus = CompetitionStatus.Finished
      competition.save()
      log.debug('Competition status updated...')
      jmsBrokerManagementService.deleteTopicsAndQueues()
      log.debug('JMS Topics and Queues purged...')
    } catch (Exception e) {
      log.warn "An Exception occurred while stopping competition ${competition}: ${e.message}"
      throw new CompetitionManagementException("An Exception occurred while stopping competition ${competition}",e)
    }
  }



  public void reset(Competition competition) throws CompetitionManagementException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot reset competition ${competition}")
    try {
      def currentCompetitionStatus = competition.competitionStatus
      stop(competition)
      if (currentCompetitionStatus == CompetitionStatus.Running) queueOutputService.sendCompetitionNotification(competition, ActionType.Reset)
      deleteCompetionDomainClasses(competition)
      initCompetition(competition)
    } catch (Exception e) {
      log.warn "An Exception occurred while resetting competition ${competition}: ${e.message}"
      throw new CompetitionManagementException("An Exception occurred while resetting competition ${competition}",e)
    }
  }

  public void delete(Competition competition) throws CompetitionManagementException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot delete competition ${competition}")
    try {
      stop(competition)
      deleteCompetionDomainClasses(competition)
      competition.delete()
    } catch (Exception e) {
      log.warn "An Exception occurred while deleting competition ${competition}: ${e.message}"
      throw new CompetitionManagementException("An Exception occurred while deleting competition ${competition}",e)
    }
  }

  /* ---------------
  * Private methods
  * ---------------
  */

  private void reScheduleProductAdvanceJob(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot reschedule competition ${competition}")
    def trigger = quartzScheduler.getTrigger('mySimpleTrigger', 'mySimpleTriggerGroup')
    trigger.repeatInterval = (competition.durationBetweenShifts * 1000l)    // in millisecods, a long value
    Calendar cal = new GregorianCalendar() //Initialized by default with "now" time
    cal.add(Calendar.SECOND, competition.initialDuration as Integer)
    log.info "First shift scheduled for: ${cal.getTime()}"
    trigger.startTime = cal.getTime()
    //trigger.repeatCount = 10           // Optional int value, if not set it repeats indefinately.
    quartzScheduler.rescheduleJob(trigger.name, trigger.group, trigger)
    quartzScheduler.start()
  }


  private void deleteCompetionDomainClasses(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot delete competition domain classes for competition ${competition}")
    Forecast.executeUpdate('delete from Forecast f where f.competition = ?', [competition])
    DepotPosition.executeUpdate('delete from DepotPosition d where d.competition = ?', [competition])
    CashPosition.executeUpdate('delete from CashPosition c where c.competition = ?', [competition])
    Orderbook.executeUpdate('delete from Orderbook o where o.competition = ?', [competition])
    Shout.executeUpdate('delete from Shout s where s.competition = ?', [competition])
    TransactionLog.executeUpdate('delete from TransactionLog t where t.competition = ?', [competition])
    Person.executeUpdate('delete from Person p where p.competition = ?', [competition])
    Product.executeUpdate('delete from Product p where p.competition = ?', [competition])
  }

  private void enableParticipants(Competition competition, boolean enable) throws PersonEnableException {
    if (!competition) throw new CompetitionNotFoundException("Cannot enable participants in competition ${competition}")
    Person.findAllByCompetition(competition)?.each {person ->
      if (enable && !person.ready) throw new PersonEnableException("Person $person is not yet ready and thus cannot be enabled.")
      person.enabled = enable
      if (!person.save()) throw new PersonEnableException("Failed to set Person.enabled to ${enable}: ${person.errors}")
    }
  }

  private void setQueueNames(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot set queue names for competition ${competition}")
    def users = Person.findAllByCompetition(competition)
    String queueNames = ""
    users.eachWithIndex { person, i ->
      if (i == 0) {
        queueNames = queueNames + person.toQueueName()
      } else {
        queueNames = queueNames + ",${person.toQueueName()}"
      }
    }
    competition.queueNames = queueNames
    competition.save()
    log.debug "Generated String containing all queueNames: ${competition.queueNames} and saved to domain competition"
  }

  private void createParticipants(Competition competition) throws PersonCreationException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot create participants for ${competition}")
    for (i in 1..competition?.participantCount) {
      def participant = new Person(username: jmsparticipants[i]['user'], userRealName: "Agent ${jmsparticipants[i]['user']}", enabled: false, passwd: jmsparticipants[i]['password'], apiKey: jmsparticipants[i]['password'])
      if(!participant.save()) throw new PersonCreationException("Failed to create person ${participant}: ${participant?.errors}")
      log.debug "Forecasts generated"
    }
  }

  private static final jmsparticipants = [
      1: [user: 'tacenergy1', password: 'secret'],
      2: [user: 'tacenergy2', password: 'secret'],
      3: [user: 'tacenergy3', password: 'secret'],
      4: [user: 'tacenergy4', password: 'secret'],
      5: [user: 'tacenergy5', password: 'secret'],
      6: [user: 'tacenergy6', password: 'secret'],
      7: [user: 'tacenergy7', password: 'secret'],
      8: [user: 'tacenergy8', password: 'secret'],
      9: [user: 'tacenergy9', password: 'secret'],
      10: [user: 'tacenergy10', password: 'secret'],
      11: [user: 'tacenergy11', password: 'secret'],
      12: [user: 'tacenergy12', password: 'secret'],
      13: [user: 'tacenergy13', password: 'secret'],
      14: [user: 'tacenergy14', password: 'secret'],
      15: [user: 'tacenergy15', password: 'secret'],
      16: [user: 'tacenergy16', password: 'secret'],
      17: [user: 'tacenergy17', password: 'secret'],
      18: [user: 'tacenergy18', password: 'secret'],
      19: [user: 'tacenergy19', password: 'secret'],
      20: [user: 'tacenergy20', password: 'secret']
  ]


}
