/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

import edu.kit.iism.cdamarket.profile.H0Generator

class ForecastGenerationService {

  boolean transactional = true


  public initForecasts(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Cannot generate forecasts for competition ${competition}.")
    Person.findAllByCompetition(competition)?.each {person ->
      generateForecastsFor(competition, person)  
    }
  }

  /**
   * Creates artificial product forecasts for all persons and all products in a competition based on their real (historic) product demands and a random walk to distort it over time
   * the algorithm itself is described in the TAC Energy concept paper at http://papers.ssrn.com/sol3/papers.cfm?abstract_id=1522683
   *
   * The basic implementation idea is as follows:
   * - Assume *all* n products for a competition (as defined in competition property 'overlallTimeslots' are created
   * - Create an hashMap "forecastMap" like this:
   *    [
   *     product_1: [trueValue: x, previousErrors: 0, counter: 0],
   *     product_2: [trueValue: y, previousErrors: 0, counter: 0],
   *     {...}
   *     product_n: [trueValue: z, previousErrors: 0, counter: 0],
   *    ]
   * - Loop backwards in time (serialNumber desc) through the product list and...
   * - ...delegate to generateForecastsFor(personX, competitiony, forecastMap, currentProduct)
   * -> Looping backwards allows my to build the random walk with incremental variability
   *
   */
  private generateForecastsFor(Competition competition, Person person) throws CompetitionNotFoundException, ProductNotFoundException {
    H0Generator timeseriesGenerator = new H0Generator()

    if (!competition) throw new CompetitionNotFoundException("Cannot generate forecasts for competition ${competition}.")
    if (!person) throw new PersonNotFoundException("Cannot generate forecasts for ${competition}. Person $person not found.")
    def products = Product.findAllByCompetition(competition, [sort: 'serialNumber', order: 'desc'])
    if (!products) throw new ProductNotFoundException("Cannot generate forecasts for ${competition}. No products found.")

    //Create map for all prodcuts first
    def forecastMap = [:]
    products.eachWithIndex {currentProduct, i -> //note: products sorted by serialNumber desc (i.e. the loop runs backwards)
      forecastMap[currentProduct] = [trueValue: timeseriesGenerator.retrieveLoad(currentProduct.startDateTime), previousErrors: 0, counter: 0]
    }

    //Now sequentially build forecasts looping "backwards in time"
    products.eachWithIndex {currentProduct, i -> //note: products sorted by serialNumber desc (i.e. the loop runs backwards)
      generateForecastsFor(person, competition, forecastMap, currentProduct)
    }


  }

  private generateForecastsFor(Person person, Competition competition, Map targetProducts, Product currentProduct) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not create forecasts for competition ${competition}.")
    Random rand = new Random(person.forecastSeed)
    targetProducts.keySet()?.each {product ->
      def forecastHorizon = product.serialNumber - currentProduct.serialNumber
      if (forecastHorizon < competition.timeslotsOpen + competition.deactivateTimeslotsAhead) { //only generate forecasts that are really needed, i.e. within the forecast horizon
        def forecast = targetProducts[product].trueValue //assume forecast = trueValue
        if (forecastHorizon > 0) {//forecastTimeslot and current timeslot different -> calculate forecast
          def counter = targetProducts[product].counter
          targetProducts[product].counter = counter + 1
          def stdBase = person.aMin * (Math.sqrt(2 * Math.PI)) / 2
          def std = stdBase * (1 + ((counter * ((person.aMax / person.aMin) - 1)) / forecastHorizon))
          def currentError = (rand.nextGaussian() * std)
          currentError += targetProducts[product].previousErrors
          targetProducts[product].previousErrors = currentError
          forecast = (1 + currentError) * targetProducts[product].trueValue
        }
        def forecastInstance = new Forecast(person: person, product: product, competitionTime: currentProduct.startDateTime, trueValue: targetProducts[product].trueValue, forecastValue: forecast.round(2))
        if (!forecastInstance.save(flush:true)) {
          log.error "Forecast cannot be saved: ${forecastInstance.errors}"
          throw new ForecastCreationException("Cannot create forecast for ${competition}, person: ${person}, product: ${product}: ${forecastInstance?.errors}")
        }
      }
    }
  }
}

/*

  def products = Product.findAllByCompetition(competition, [sort: 'serialNumber', order: 'asc'])

  def current = products.last()

  def forecastHorizon = targetTimeslot.serialNumber - currentTimeslot.serialNumber

  new Forecast(Person: person,
      product: targetTimeslot,
      forecastValue: realValue, //if currentTimeslot = targetTimeslot -> forecastError = 0
      trueValue: realValue
  ).save() //forecast for the targetTimeslot at the startTime of the targetTimeslot is the true realValue
  int i = 0
  while (current.serialNumber > currentTimeslot.serialNumber) { //increase i with each step but go "backwards in time from targetTimeslot (forecastError = 0) to currentTimeslot (forecastError = max)
    i++
    current = current.previous()
    if (current) {
      def forecast = realValue
      if (aMin > 0 && aMax > 0) {
        def std = stdBase * (forecastHorizon + (i * aMax / aMin) - i - (aMax / aMin)) / (1 + Math.sqrt(1 + (alpha * alpha * i) - (alpha * alpha)) * (forecastHorizon - 1))
        def currentError = (rand.nextGaussian() * std)
        currentError += (alpha * previousErrors)
        previousErrors += currentError
        forecast = (1 + currentError) * realValue
      }
      new Forecast(Person: person,
          product: targetTimeslot,
          forecastValue: forecast.round(2),
          trueValue: realValue
      ).save() //forecast for the "now" targetTimeslot is the true realValue
    }
  }
}
*/
