/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

import groovy.time.TimeDuration
import java.text.SimpleDateFormat

class ProductManagementService {

  boolean transactional = true

  def balancingPowerService
  def queueOutputService
  def idGeneratorService
  def quartzScheduler

  SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("dd.MM.yy-HH:mm:ss:SSS")
  SimpleDateFormat SHORT_DATE_FORMAT = new SimpleDateFormat("dd.MM HH:mm")

  //"local" class variables required to build the timeslotsOverall
  private TimeDuration slotDuration //is typically one second less than the intended timeslot, e.g 14 minutes 59 seconds for a "15 min" timeslot
  private TimeDuration slotDistance //describes the duration between two timeslot startDateTimes, e.g. 15 minutes for a "15 min" timeslot

  /*
  * +----------------+
  * | Public methods |
  * +----------------+
  */

  /**
   * Deletes all existing products (and implicitly also all dependent shout and depot positions) and creates
   * an initial set of <code>overallProducts</code> prodcuts, which are all disabled
   * @return List of all created products
   */
  public List initProducts(Competition competition) throws ProductCreationException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not create new product, corresponding competition was ${competition}.")

    //cleanup existing products in this competition -> triggers deletion of dependent depot positions and shouts
    def productList = Product.findAllByCompetition(competition)
    productList?.each {
      it.delete()
    }

    if (!slotDuration) slotDuration = new TimeDuration(0, competition.timeslotLength - 1, 59, 0) //1 second less than the defined duration from Product domain class)
    if (!slotDistance) slotDistance = new TimeDuration(0, competition.timeslotLength, 0, 0)
    GregorianCalendar cal = new GregorianCalendar()
    cal.setTime(competition.simulationStartTime)
    cal.set(Calendar.MINUTE, 0)
    cal.set(Calendar.SECOND, 0)
    cal.set(Calendar.MILLISECOND, 0)

    Date startTime = cal.getTime()
    Date endTime = slotDuration.plus(startTime)
    Boolean createNextTimeslot = true
    for (long i in 1..competition.timeslotsOverall) {
      def productInstance = new Product(name: "[${SHORT_DATE_FORMAT.format(startTime)} - ${SHORT_DATE_FORMAT.format(endTime)}]", serialNumber: i, startDateTime: startTime, endDateTime: endTime, enabled: false, current: (i == 1)) //competition is automatically injected before saving
      if (!productInstance.save()) throw new ProductCreationException("Failed to create product ${productInstance}: ${productInstance?.errors}")
      startTime = slotDistance.plus(startTime)
      endTime = slotDistance.plus(endTime)
      if (log.isInfoEnabled()) log.info "New timeslot '$productInstance' created during timeslot initialization."
      productList.add(productInstance)
    }
    return productList
  }

  public openInitialProducts(Competition competition) throws NoCurrentProductException, OpenInitialProductsException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not open initial products, corresponding competition was ${competition}.")
    def products = getAllProducts(competition)
    if (!products) throw new OpenInitialProductsException("Could not open initial products: ${products?.size()} found.")
    def currentProduct = getCurrentProduct(competition)
    if (!currentProduct) throw new NoCurrentProductException("Cannot open initial products for competition $competition: No current product found.")
    products.eachWithIndex {product, i ->
      def timeAhead = competition.deactivateTimeslotsAhead
      // send product notifications for deactivated products at the start of the competition
      if (i < timeAhead) {
        queueOutputService.sendProductNotification(product, ActionType.Stopped)
        log.info("inital deactivated product ${product} annonunced")
      }
      if (i >= timeAhead && i < (timeAhead + competition.timeslotsOpen)) {
        product.enabled = true
        if (!product.save()) throw new OpenInitialProductsException("Could enable product '${product}': ${product?.errors}")
        queueOutputService.sendProductNotification(product, ActionType.Started)
        log.info("Product opened: ${product}")
      }
    }
    competition.currentCompetitionTime = currentProduct.startDateTime
    competition.save()
    queueOutputService.sendCompetitionNotification(competition, ActionType.TimeAdvanced)
    log.info("New competition time announced: ${currentProduct.startDateTime}")
    log.debug('Trying to send out new forecasts...')
    sendProductForecasts(competition)
    return null
  }

  public shiftProducts() throws CompetitionNotFoundException {
    try {
      def competition = idGeneratorService.competition
      if (!competition) throw new CompetitionNotFoundException("Could not shift products for competition ${competition}.")
      shiftProducts(competition)
    } catch (Exception e) {
      log.error("Error shifting products ${e.message}: ${e.cause}", e)
    }
    return null
  }

  public shiftProducts(Competition competition) {
    log.debug "Product shift started"
    try {
      log.debug("Trying to activate new product...")
      activateNewProduct(competition)
      log.debug("Trying to deactivate old product...")
      def oldestProduct = deactivateOldestProduct(competition)
      log.debug("Trying to advance 'current' product market...")
      advanceCurrentProduct(competition)
      log.debug('Trying to send out new forecasts...')
      sendProductForecasts(competition)
      log.debug('Calculating and billing balancing power demand for deactivated product')
      balancingPowerService.balance(oldestProduct)
    } catch (NoActiveProductException nape) {
      log.error("No active products found -> Competition end reached.")
      queueOutputService.sendCompetitionNotification(competition, ActionType.Stopped)
      log.debug('Competition stop announced...')
      quartzScheduler.pauseTrigger('mySimpleTrigger', 'mySimpleTriggerGroup')
      log.debug('Quartz scheduler stopped...')
      Person.findAllByCompetition(competition)?.each {person ->
        person.enabled = false
        person.save(flush: true)
      }
      log.debug('Participants disabled...')
      Competition.executeUpdate('update Competition c set c.enabled=false, competitionStatus=? where c.id=?', [CompetitionStatus.Finished, competition.id])
      log.debug('Competition status updated...')
    } catch (Exception e) {
      log.error("Error shifting products ${e.message}: ${e.cause}", e)
    }
    log.debug('Product shift completed')
    return null
  }

  public advanceCurrentProduct(Competition competition) throws ProductShiftException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not advance current product for competition ${competition}.")
    def currentProduct = getCurrentProduct(competition)
    if (!currentProduct) throw new ProductShiftException("Cannot advance current product for competition $competition: No current product found.")
    def newCurrentProduct = currentProduct.next()
    if (newCurrentProduct) {
      currentProduct.current = false
      if (!currentProduct.save(flush: true)) throw new ProductShiftException("Failed to update current property for (old) current product ${currentProduct}: ${currentProduct?.errors}")
      log.debug("Product ${currentProduct} no longer current product")
      newCurrentProduct.current = true
      if (!newCurrentProduct.save(flush: true)) throw new ProductShiftException("Failed to update current property for (new) current product ${newCurrentProduct}: ${newCurrentProduct?.errors}")
      log.debug("Product ${newCurrentProduct} now current product")
      competition.currentCompetitionTime = newCurrentProduct.startDateTime
      competition.save()
      queueOutputService.sendCompetitionNotification(competition, ActionType.TimeAdvanced)
      log.info("New competition time: ${currentProduct.startDateTime}")
    }
  }

  /**
   * Deactivates the oldest still enabled product (i.e. in tac energy timeslot)
   * @return returns the oldest formerly active timeslot after its deactivation
   */
  public Product deactivateOldestProduct(Competition competition) throws NoActiveProductException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not deactivate oldest product, corresponding competition was ${competition}.")
    def oldestProduct = Product.findByCompetitionAndEnabled(competition, true, [sort: 'serialNumber', order: 'asc'])
    if (!oldestProduct) throw new NoActiveProductException("Could not deactivate oldest product. No active products found for competition $competition")
    oldestProduct = switchProductEnabled(competition, oldestProduct, false)
    if (log.isInfoEnabled()) log.info "Product '$oldestProduct' disabled"
    return oldestProduct
  }

  public Product activateNewProduct(Competition competition) throws NoActiveProductException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not activate new product, corresponding competition was ${competition}.")
    def newestProduct = Product.findByCompetitionAndEnabled(competition, true, [sort: 'serialNumber', order: 'desc'])
    if (!newestProduct) throw new NoActiveProductException("Could not activate new products. No currently active products found for competition $competition")
    def productToEnable = newestProduct.next()
    if (productToEnable) {
      if (log.isInfoEnabled()) log.info "Product '$productToEnable' enabled"
      return switchProductEnabled(competition, productToEnable, true)
    } else {
      if (log.isInfoEnabled()) log.info "No product after '$newestProduct' found to enable..."
      return null // no new (not yet activated) product found after newest already activated product
    }
  }

  /**
   * Sets enabled flag for the given product (i.e. in tac energy timeslot)
   * @param competition current competition
   * @param product product to adapt
   * @param enable true or false to enable / disable product
   * @return returns the given product after its status was changed, the change was persisted to the db and the status notification was send to participants via jms
   */
  public Product switchProductEnabled(Competition competition, Product product, Boolean enable) throws ProductEnableDisableException, ProductNotFoundException, CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not deactivate product, corresponding competition was ${competition}.")
    if (!product) throw new ProductNotFoundException("Could not deactivate product '${product}'")
    product.enabled = enable ?: false
    if (!product.save(flush: true)) throw new ProductEnableDisableException("Failed to set product.enabled to ${enable} for ${product}: ${product?.errors}")
    queueOutputService.sendProductNotification(product, enable ? ActionType.Started : ActionType.Stopped)
    return product
  }

  public List getAllProducts(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not retrieve products for competition ${competition}")
    return Product.findAllByCompetition(competition, [sort: 'serialNumber', order: 'asc'])
  }

  public List getOpenProducts(Competition competition) throws CompetitionNotFoundException {
    return getOpenProducts(competition, null)
  }

  public List getOpenProducts(Competition competition, Integer max) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not retrieve open products for competition ${competition}")
    return Product.withCriteria {
      eq('competition', competition)
      eq('enabled', true)
      if (max) maxResults(max)
      order('serialNumber', 'asc')
    }
  }

  public Product getCurrentProduct(Competition competition) throws CompetitionNotFoundException {
    if (!competition) throw new CompetitionNotFoundException("Could not retrieve current product for competition ${competition}")
    Product.findByCompetitionAndCurrent(competition, true)
  }

  public void sendProductForecasts(Competition competition) {
    if (!competition) throw new CompetitionNotFoundException("Could not send out product forecasts for for competition ${competition}")
    def currentProduct = getCurrentProduct(competition)
    if (!currentProduct) throw new NoCurrentProductException("Could not send out product forecasts for for competition ${competition}. No current product found.")

    // nextEnabeledTimeslot is the serial number of the next enabled product in the future
    def nextEnabledTimeslot = currentProduct.serialNumber + competition.deactivateTimeslotsAhead

    // Find all forecast starting with the next enabled product in the future.
    // In consequence to the generation process these are only the forecasts for all enabled products.
    def forecastList = Forecast.withCriteria {
      and {
        eq('competition', competition)
        eq('competitionTime', currentProduct.startDateTime)
        product {
          ge('serialNumber', nextEnabledTimeslot)
        }  
      }
    }

    forecastList?.each {forecast ->
      queueOutputService.sendForecast(forecast)
    }

  }
}
