/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

import grails.plugin.jms.Queue
import groovy.util.slurpersupport.GPathResult
import org.springframework.security.BadCredentialsException
import org.springframework.security.DisabledException

class QueueManagementService {

  def auctionService
  def queueOutputService
  def idGeneratorService

  boolean transactional = true
  static expose = ['jms']

  @Queue(name = "public.ShoutInputQueue")
  def receiveShoutNotification(String xmlString) throws CompetitionNotFoundException, NullPointerException {
    log.debug("Received new shout:  ${xmlString}")

    def competition = idGeneratorService.competition
    if (!competition) throw new CompetitionNotFoundException("Competition ${competition} not found while processing order ${xmlString}")
    if (!xmlString) throw new NullPointerException("Cannout process incoming shout ${xmlString}")

    def xml
    try {
      xml = new XmlSlurper().parseText(xmlString)
      def person = checkAuthorizationIgnoreEnabled(competition, xml.username?.text(), xml.apikey?.text())

      if (xml.action.text().equalsIgnoreCase('Create')) {
        processNewIncomingShout(competition, person, xml)
      } else if (xml.action.text().equalsIgnoreCase('Update')) {
        processShoutUpdate(competition, person, xml)
      } else if (xml.action.text().equalsIgnoreCase('Delete')) {
        processShoutDelete(competition, person, xml)
      } else {
        throw new OrderEntryException("Failed to process shout: Failed to process 'action' property ${xml.action}")
      }

    } catch (OrderEntryException e) {
      log.error("An error occurred while processing the incoming shout ${xmlString}: ${e.message}", e)
      try {
        sendQueueJMSMessage("users.${xml.username?.text()}.error", e.message)
      } catch (Exception sendException) {
        log.error("Failed to send error message to queue users.${xml.username?.text()}.error", sendException)
      }
    } catch (DisabledException de) {
      log.error("A disabled user tried submit a shout: ${de.message}")
      try {
        sendQueueJMSMessage("users.${xml.username?.text()}.error", de.message)
      } catch (Exception sendException) {
        log.error("Failed to send error message to queue users.${xml.username?.text()}.error", sendException)
      }
    } catch (Exception ex) {
      log.error("Error creating new shout: ${ex.message}", ex)
    }
    return null
  }

  private void processNewIncomingShout(Competition competition, Person person, GPathResult xml) throws OrderEntryException, MarketOrderException, ExecutionException, SettlementException, ShoutNotFoundException {
    def productId = xml.productId?.text()
    if (!productId) throw new OrderEntryException("Failed to create new shout: Failed to retrieve product ${productId}.")
    def product = Product.get(productId)
    if (!product) throw new OrderEntryException("Failed to create new shout: No product found for productId: ${productId}")
    String buySellIndicatorTemp = xml.buySellIndicator.text()
    if (!buySellIndicatorTemp) throw new OrderEntryException("Failed to create new shout: Property 'buySellIndicator' missing.")
    BuySellIndicator buySellIndicator = null
    try {buySellIndicator = buySellIndicatorTemp as BuySellIndicator} catch (Exception e) {}
    if (!buySellIndicator) throw new OrderEntryException("Failed to create new shout: Failed to parse buySellIndicator '${buySellIndicatorTemp}'. Invalid value.")
    def remainingQuantityTemp = xml.remainingQuantity?.text()
    if (remainingQuantityTemp == null) throw new OrderEntryException("Failed to create new shout: Property 'remainingQuantity' missing.")
    BigDecimal remainingQuantity = null
    try {remainingQuantity = remainingQuantityTemp as BigDecimal} catch (Exception e) {}
    if (remainingQuantity == null) throw new OrderEntryException("Failed to create new shout: Failed to parse remainingQuantity '${remainingQuantityTemp}'. Invalid value.")
    def orderTypeTemp = xml.orderType?.text()
    if (orderTypeTemp == null) throw new OrderEntryException("Failed to create new shout: Property 'orderType' missing.")
    OrderType orderType = null
    try {orderType = (orderTypeTemp as OrderType)} catch (Exception e) {}
    if (orderType == null) throw new OrderEntryException("Failed to create new shout: Failed to parse orderType '${orderType}'. Invalid value.")
    BigDecimal limitPrice = null
    if (orderType == OrderType.LIMIT) {
      def limitPriceTemp = xml.limitPrice?.text()
      if (limitPriceTemp == null) throw new OrderEntryException("Failed to create new shout: Property 'limitPrice' missing.")
      try {limitPrice = limitPriceTemp as BigDecimal} catch (Exception e) {}
      if (limitPrice == null) throw new OrderEntryException("Failed to create new shout: Failed to parse limitPrice '${limitPrice}'. Invalid value.")
    }

    def shoutInstance = new Shout()
    shoutInstance.person = person
    shoutInstance.competition = competition
    shoutInstance.product = product
    shoutInstance.buySellIndicator = buySellIndicator
    shoutInstance.remainingQuantity = remainingQuantity
    shoutInstance.limitPrice = limitPrice
    shoutInstance.orderType = orderType

    def createdShoutInstance = auctionService.addShout(shoutInstance)
    queueOutputService.sendOrderstatus(createdShoutInstance)
    shoutInstance.discard()
  }


 private void processShoutUpdate(Competition competition, Person person, GPathResult xml) throws OrderUpdateException, MarketOrderException, ExecutionException, SettlementException, ShoutNotFoundException {
    def shoutId = xml.shoutID?.text()
    if (!shoutId) throw new OrderUpdateException("Failed to update shout: No shout id found: ${shoutId}.")
    def shoutInstance = Shout.findByShoutID(shoutId)
    if (!shoutInstance) throw new OrderUpdateException("Failed to update shout: No shout found for id: ${shoutId}")
    def remainingQuantityTemp = xml.remainingQuantity?.text()
    if (remainingQuantityTemp == null) throw new OrderUpdateException("Failed to update shout: Property 'remainingQuantity' missing.")
    BigDecimal remainingQuantity = null
    try {remainingQuantity = remainingQuantityTemp as BigDecimal} catch (Exception e) {}
    if (remainingQuantity == null) throw new OrderUpdateException("Failed to update shout: Failed to parse remainingQuantity '${remainingQuantityTemp}'. Invalid value.")
    def orderTypeTemp = xml.orderType?.text()
    if (orderTypeTemp == null) throw new OrderUpdateException("Failed to update shout: Property 'orderType' missing.")
    OrderType orderType = null
    try {orderType = orderTypeTemp as OrderType} catch (Exception e) {}
    if (orderType == null) throw new OrderUpdateException("Failed to update shout: Failed to parse orderType '${orderType}'. Invalid value.")
    BigDecimal limitPrice = null
    if (orderType == OrderType.LIMIT) {
      def limitPriceTemp = xml.limitPrice?.text()
      if (limitPriceTemp == null) throw new OrderUpdateException("Failed to update shout: Property 'limitPrice' missing.")
      try {limitPrice = limitPriceTemp as BigDecimal} catch (Exception e) {}
      if (limitPrice == null) throw new OrderUpdateException("Failed to update shout: Failed to parse limitPrice '${limitPrice}'. Invalid value.")
    }

    shoutInstance.remainingQuantity = remainingQuantity
    shoutInstance.limitPrice = limitPrice
    shoutInstance.orderType = orderType

    def updatedShoutInstance = auctionService.editShout(shoutInstance)
    queueOutputService.sendOrderstatus(updatedShoutInstance)
    shoutInstance.discard()
  }

  private void processShoutDelete(Competition competition, Person person, GPathResult xml) throws OrderDeleteException, MarketOrderException, ExecutionException, SettlementException, ShoutNotFoundException {
    def shoutId = xml.shoutID?.text()
    if (!shoutId) throw new OrderDeleteException("Failed to delete shout: No shout id found: ${shoutId}.")
    def shoutInstance = Shout.findByShoutID(shoutId)
    if (!shoutInstance) throw new OrderDeleteException("Failed to delete shout: No shout found for id: ${shoutId}")
    
    def invalidatedShoutInstance = auctionService.deleteShout(shoutInstance)
    queueOutputService.sendOrderstatus(invalidatedShoutInstance)
    shoutInstance.discard()
  }

  @Queue(name = "public.ClientAnnounce")
  def receiveClientAnnounce(xmlString) throws CompetitionNotFoundException {
    def competition = idGeneratorService.competition
    if (!competition) throw new CompetitionNotFoundException("Competition ${competition} not found while processing order ${incomingShoutMap}")

    try {
      def xml = new XmlSlurper().parseText(xmlString)
      def person = checkAuthorizationIgnoreEnabled(competition, xml.username?.text(), xml.apikey?.text())
      person?.ready = true
      if (!person?.save(flush: true)) throw new PersonEnableException("Failed to enable person ${person}: ${person.errors}")
    } catch (Exception e) {
      log.error("Error processing client notificiation ${xmlString}", e)
    }
    return null
  }

  private Person checkAuthorization(Competition competition, String userName, String apiKey) throws DisabledException, BadCredentialsException {
    if (!userName || !apiKey) throw new BadCredentialsException("Cannot authorize user ${userName}: Name or apiKey was null.")
    def person = Person.withCriteria(uniqueResult: true) {
      eq('username', userName)
      eq('apiKey', apiKey)
      eq('competition', competition)
    }
    if (person) {
      if (person.enabled) {
        return person
      } else {
        throw new DisabledException("Cannot authorize user ${userName}: Account disabled.")
      }
    } else {
      throw new BadCredentialsException("Cannot authorize user ${userName}: Wrong or missing username or apiKey")
    }
  }

  private Person checkAuthorizationIgnoreEnabled(Competition competition, String userName, String apiKey) throws BadCredentialsException {
    if (!userName || !apiKey) throw new BadCredentialsException("Cannot authorize user ${userName}: Name or apiKey was null.")
    def person = Person.withCriteria(uniqueResult: true) {
      eq('username', userName)
      eq('apiKey', apiKey)
      eq('competition', competition)
    }
    if (person) {
      return person
    } else {
      throw new BadCredentialsException("Cannot authorize user ${userName}: Wrong or missing username or apiKey")
    }
  }
}
