/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.cdamarket

import grails.test.GrailsUnitTestCase

class AuctionServiceIntegrationTests extends GrailsUnitTestCase {

  /*GENERAL NOTE:
   * - transactionIds are assumed to start at 1 for each test case though in the production system they will be determined by a transactionId Generator within the auctionService
   * - Note that the transactionIds are identical for all shout, transactionLog, and ordebook entries that were manipulated within a single transaction
   * - shoutIds are assumed also assumed to start at 1 for each test case though in the production system they will be determined by a shoutId Generator within the Shout domain class
   * - Note that shoutIds are identitcal throughout all (cloned) instances of a shout while the doman classes "id" property is not
   * - In all of the following tests "shout.outdated" has to be set to false only for the latest cloned instance of a shout while it has to be set to true for all "predecessor" instances
   */

  def competition
  def sampleProduct
  def sampleSeller
  def sampleBuyer
  Shout sell1
  Shout sell2
  Shout sell3
  Shout sell4
  Shout buy1
  Shout buy2
  Shout buy3
  Shout buy4
  Shout msell
  Shout mbuy
  def auctionService

  protected void setUp() {
    super.setUp()
    competition = new Competition(name: "test", enabled: true)
    competition.timeslotsOverall = 5
    competition.timeslotLength = 15
    competition.timeslotsOpen = 3
    competition.deactivateTimeslotsAhead = 1
    competition.simulationStartTime = new GregorianCalendar(2006, 05, 05, 14, 00).getTime()
    competition.current = true
    assert(competition.save())
    sampleSeller = new Person(username: "SampleSeller", userRealName: 'SampleSeller', passwd:'Secret', email: 'testseller@test.com', enabled: true)
    assert(sampleSeller.save())
    sampleBuyer = new Person(username: "SampleBuyer", userRealName: 'SampleBuyer', passwd:'Secret', email: 'testbuyer@test.com', enabled: true)
    assert(sampleBuyer.save())
    sampleProduct = new Product(name: "SampleProduct", enabled: true, serialNumber: 1L)
    assert(sampleProduct.save())
    sell1 = new Shout(person: sampleSeller, product: sampleProduct, buySellIndicator: BuySellIndicator.SELL, orderType: OrderType.LIMIT)
    sell2 = new Shout(person: sampleSeller, product: sampleProduct, buySellIndicator: BuySellIndicator.SELL, orderType: OrderType.LIMIT)
    sell3 = new Shout(person: sampleSeller, product: sampleProduct, buySellIndicator: BuySellIndicator.SELL, orderType: OrderType.LIMIT)
    sell4 = new Shout(person: sampleSeller, product: sampleProduct, buySellIndicator: BuySellIndicator.SELL, orderType: OrderType.LIMIT)
    buy1 = new Shout(person: sampleBuyer, product: sampleProduct, buySellIndicator: BuySellIndicator.BUY, orderType: OrderType.LIMIT)
    buy2 = new Shout(person: sampleBuyer, product: sampleProduct, buySellIndicator: BuySellIndicator.BUY, orderType: OrderType.LIMIT)
    buy3 = new Shout(person: sampleBuyer, product: sampleProduct, buySellIndicator: BuySellIndicator.BUY, orderType: OrderType.LIMIT)
    buy4 = new Shout(person: sampleBuyer, product: sampleProduct, buySellIndicator: BuySellIndicator.BUY, orderType: OrderType.LIMIT)
    msell = new Shout(person: sampleSeller, product: sampleProduct, buySellIndicator: BuySellIndicator.SELL, orderType: OrderType.MARKET)
    mbuy = new Shout(person: sampleBuyer, product: sampleProduct, buySellIndicator: BuySellIndicator.BUY, orderType: OrderType.MARKET)
  }

  protected void tearDown() {
    super.tearDown()
  }

  /**
   * Init: New Sell 10@10EUR
   * Action: New Sell 1@10EUR
   * Validate:
   * 1.) 2 shout objects in db
   * - Sell 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Sell 1@10EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   *
   * 2.) 2 TransactionLog entries:
   * - Quote: Ask: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
   * - Quote: Ask: 10, AskSize: 11, Bid: null, BidSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - ask1: 10 EUR, askSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - ask1: 10 EUR, askSize1: 11, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewSellOrderEntryCreatesNewQuoteWithNewQuantity() {
    //init
    sell1.limitPrice = 10.0
    sell1.remainingQuantity = 10.0
    auctionService.addShout(sell1)
    //action
    sell2.limitPrice = 10.0
    sell2.remainingQuantity = 1.0
    auctionService.addShout(sell2)

    /* 1.) 2 shout objects in db
   * - Sell 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Sell 1@10EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   */
    assertEquals(2, Shout.list().size())

    Shout.list().each {shout ->
      assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
    }

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 10.0)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    def s2 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 1.0)
    }
    assertNotNull s2
    def tid2 = s2.transactionID
    def sid2 = s2.shoutID

    /* 2.) 2 TransactionLog entries:
    * - Quote: Ask: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
    * - Quote: Ask: 10, AskSize: 11, Bid: null, BidSize: 0, transactionId: 2
    */
    assertEquals(2, TransactionLog.list().size())
    TransactionLog tl1 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('ask', 10.0)
      eq('askSize', 10.0)
      eq('bidSize', 0.0)
      isNull('bid')
    }
    assertNotNull tl1
    assertEquals(tid1, tl1.transactionID)

    TransactionLog tl2 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('ask', 10.0)
      eq('askSize', 11.0)
      eq('bidSize', 0.0)
      isNull('bid')
    }
    assertNotNull tl2
    assertEquals(tid2, tl2.transactionID)

    /* 3.) 2 Orderbook entries:
    * - ask1: 10 EUR, askSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
    * - ask1: 10 EUR, askSize1: 11, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
    */
    assertEquals(2, Orderbook.list().size())

    Orderbook ob1 = Orderbook.withCriteria(uniqueResult: true) {
      eq('ask1', 10.0)
      eq('askSize1', 10.0)
      eq('bidSize1', 0.0)
      eq('askSize2', 0.0)
      isNull('ask2')
      isNull('bid1')
    }
    assertNotNull ob1
    assertEquals(tid1, ob1.transactionID)

    Orderbook ob2 = Orderbook.withCriteria(uniqueResult: true) {
      eq('ask1', 10.0)
      eq('askSize1', 11.0)
      eq('bidSize1', 0.0)
      eq('askSize2', 0.0)
      isNull('ask2')
      isNull('bid1')
    }
    assertNotNull ob2
    assertEquals(tid2, ob2.transactionID)
  }

  /**
   * Init: New Buy 10@10EUR
   * Action: New Buy 1@10EUR
   * Validate:
   * 1.) 2 shout objects in db
   * - Buy 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Buy 1@10EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   *
   * 2.) 2 TransactionLog entries:
   * - Quote: Bid: 10, BidSize: 10, Ask: null, AskSize: 0, transactionId: 1
   * - Quote: Bid: 10, BidSize: 11, Ask: null, AskSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - bid1: 10 EUR, bidSize1: 11, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewBuyOrderEntryCreatesNewQuoteWithNewQuantity() {
    //init
    buy1.limitPrice = 10.0
    buy1.remainingQuantity = 10.0
    auctionService.addShout(buy1)
    //action
    buy2.limitPrice = 10.0
    buy2.remainingQuantity = 1.0
    auctionService.addShout(buy2)

    /* 1.) 2 shout objects in db
    * - Buy 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
    * - Buy 1@10EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
    */
    assertEquals(2, Shout.list().size())

    Shout.list().each {shout ->
      assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
    }

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 10.0)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    def s2 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 1.0)
    }
    assertNotNull s2
    def tid2 = s2.transactionID
    def sid2 = s2.shoutID

    /* 2.) 2 TransactionLog entries:
    * - Quote: Bid: 10, BidSize: 10, Ask: null, AskSize: 0, transactionId: 1
    * - Quote: Bid: 10, BidSize: 11, Ask: null, AskSize: 0, transactionId: 2
    */
    assertEquals(2, TransactionLog.list().size())

    TransactionLog tl1 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('bid', 10.0)
      eq('bidSize', 10.0)
      eq('askSize', 0.0)
      isNull('ask')
    }
    assertNotNull tl1
    assertEquals(tid1, tl1.transactionID)

    TransactionLog tl2 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('bid', 10.0)
      eq('bidSize', 11.0)
      eq('askSize', 0.0)
      isNull('ask')
    }
    assertNotNull tl2
    assertEquals(tid2, tl2.transactionID)

    /* 3.) 2 Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - bid1: 10 EUR, bidSize1: 11, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
    assertEquals(2, Orderbook.list().size())

    Orderbook ob1 = Orderbook.withCriteria(uniqueResult: true) {
      eq('bid1', 10.0)
      eq('bidSize1', 10.0)
      eq('askSize1', 0.0)
      eq('bidSize2', 0.0)
      isNull('bid2')
      isNull('ask1')
    }
    assertNotNull ob1
    assertEquals(tid1, ob1.transactionID)

    Orderbook ob2 = Orderbook.withCriteria(uniqueResult: true) {
      eq('bid1', 10.0)
      eq('bidSize1', 11.0)
      eq('askSize1', 0.0)
      eq('bidSize2', 0.0)
      isNull('bid2')
      isNull('ask1')
    }
    assertNotNull ob2
    assertEquals(tid2, ob2.transactionID)
  }

  // ------------------------------------------------------------------------------------------------------------------

  /**
   * Init: New Sell 10@10EUR
   * Action: New Sell 1@9.99EUR
   * Validate:
   * 1.) 2 shout objects in db
   * - Sell 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Sell 1@9.99EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   *
   * 2.) 2 TransactionLog entries:
   * - Quote: Ask: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
   * - Quote: Ask: 9.99, AskSize: 1, Bid: null, BidSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - ask1: 10 EUR, askSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - ask1: 9.99EUR, askSize1: 1, ask2: 10 EUR, askSize2: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewSellOrderEntryCreatesNewQuoteWithNewPriceAndQuantity() {
    //init
    sell1.limitPrice = 10.0
    sell1.remainingQuantity = 10.0
    auctionService.addShout(sell1)
    //action
    sell2.limitPrice = 9.99
    sell2.remainingQuantity = 1.0
    auctionService.addShout(sell2)
    /* 1.) 2 shout objects in db
    * - Sell 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
    * - Sell 1@9.99EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
    */
    assertEquals(2, Shout.list().size())

    Shout.list().each {shout ->
      assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
    }

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 10.0)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    Shout s2 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 9.99)
      eq('remainingQuantity', 1.0)
    }
    assertNotNull s2
    def tid2 = s2.transactionID
    def sid2 = s2.shoutID

    /*2.) 2 TransactionLog entries:
    * - Quote: Ask: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
    * - Quote: Ask: 9.99, AskSize: 1, Bid: null, BidSize: 0, transactionId: 2
    */
    assertEquals(2, TransactionLog.list().size())
    TransactionLog tl1 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('ask', 10.0)
      eq('askSize', 10.0)
      eq('bidSize', 0.0)
      isNull('bid')
    }
    assertNotNull tl1
    assertEquals(tid1, tl1.transactionID)

    TransactionLog tl2 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('ask', 9.99)
      eq('askSize', 1.0)
      eq('bidSize', 0.0)
      isNull('bid')
    }
    assertNotNull tl2
    assertEquals(tid2, tl2.transactionID)
    /* 3.) 2 Orderbook entries:
    * - ask1: 10 EUR, askSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
    * - ask1: 9.99EUR, askSize1: 1, ask2: 10 EUR, askSize2: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
    */
    assertEquals(2, Orderbook.list().size())

    Orderbook ob1 = Orderbook.withCriteria(uniqueResult: true) {
      eq('ask1', 10.0)
      eq('askSize1', 10.0)
      eq('bidSize1', 0.0)
      eq('askSize2', 0.0)
      isNull('ask2')
      isNull('bid1')
    }
    assertNotNull ob1
    assertEquals(tid1, ob1.transactionID)

    Orderbook ob2 = Orderbook.withCriteria(uniqueResult: true) {
      eq('ask1', 9.99)
      eq('ask2', 10.0)
      eq('askSize1', 1.0)
      eq('askSize2', 10.0)
      eq('bidSize1', 0.0)
      eq('askSize3', 0.0)
      isNull('ask3')
      isNull('bid1')
    }
    assertNotNull ob2
    assertEquals(tid2, ob2.transactionID)
  }

  /**
   * Init: New Buy 10@10EUR
   * Action: New Buy 1@10.01EUR
   * Validate:
   * 1.) 2 shout objects in db
   * - Buy 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Sell 1@10.01EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   *
   * 2.) 2 TransactionLog entries:
   * - Quote: Bid: 10, BidSize: 10, Ask: null, AskSize: 0, transactionId: 1
   * - Quote: Bid: 10.01, BidSize: 1, Ask: null, AskSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - bid1: 10.01EUR, bidSize1: 1, bid2: 10 EUR, bidSize2: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewBuyOrderEntryCreatesNewQuoteWithNewPriceAndQuantity() {
    //init
    buy1.limitPrice = 10.0
    buy1.remainingQuantity = 10.0
    auctionService.addShout(buy1)
    //action
    buy2.limitPrice = 10.01
    buy2.remainingQuantity = 1.0
    auctionService.addShout(buy2)

    /* 1.) 2 shout objects in db
    * - Buy 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
    * - Sell 1@10.01EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
    */
    assertEquals(2, Shout.list().size())

    Shout.list().each {shout ->
      assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
    }

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 10.0)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    def s2 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.01)
      eq('remainingQuantity', 1.0)
    }
    assertNotNull s2
    def tid2 = s2.transactionID
    def sid2 = s2.shoutID

    /* 2.) 2 TransactionLog entries:
    * - Quote: Bid: 10, BidSize: 10, Ask: null, AskSize: 0, transactionId: 1
    * - Quote: Bid: 10.01, BidSize: 1, Ask: null, AskSize: 0, transactionId: 2
    */
    assertEquals(2, TransactionLog.list().size())

    TransactionLog tl1 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('bid', 10.0)
      eq('bidSize', 10.0)
      eq('askSize', 0.0)
      isNull('ask')
    }
    assertNotNull tl1
    assertEquals(tid1, tl1.transactionID)

    TransactionLog tl2 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('bid', 10.01)
      eq('bidSize', 1.0)
      eq('askSize', 0.0)
      isNull('ask')
    }
    assertNotNull tl2
    assertEquals(tid2, tl2.transactionID)

    /* 3.) 2 Orderbook entries:
    * - bid1: 10 EUR, bidSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
    * - bid1: 10.01EUR, bidSize1: 1, bid2: 10 EUR, bidSize2: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
    */

    assertEquals(2, Orderbook.list().size())

    Orderbook ob1 = Orderbook.withCriteria(uniqueResult: true) {
      eq('bid1', 10.0)
      eq('bidSize1', 10.0)
      eq('askSize1', 0.0)
      eq('bidSize2', 0.0)
      isNull('bid2')
      isNull('ask1')
    }
    assertNotNull ob1
    assertEquals(tid1, ob1.transactionID)

    Orderbook ob2 = Orderbook.withCriteria(uniqueResult: true) {
      eq('bid1', 10.01)
      eq('bidSize1', 1.0)
      eq('bid2', 10.0)
      eq('bidSize2', 10.0)
      eq('askSize1', 0.0)
      eq('bidSize3', 0.0)
      isNull('bid3')
      isNull('ask1')
    }
    assertNotNull ob2
    assertEquals(tid2, ob2.transactionID)
  }

  // -----------------------------------------------------------------------------------------------------------------

  /**
   * Init: New Sell 10@10EUR
   * Action: New Sell 1@10.01EUR
   * Validate:
   * 1.) 2 shout objects in db
   * - Sell 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Sell 1@10.01EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   *
   * 2.) 1 TransactionLog entry:
   * - Quote: Ask: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
   *
   * 3.) 2 Orderbook entries:
   * - ask1: 10 EUR, askSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - ask1: 10EUR, askSize1: 10, ask2: 10.01 EUR, askSize2: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewSellOrderEntryCreatesNewOrderbookOnly() {
    //init
    sell1.limitPrice = 10.0
    sell1.remainingQuantity = 10.0
    auctionService.addShout(sell1)
    //action
    sell2.limitPrice = 10.01
    sell2.remainingQuantity = 1.0
    auctionService.addShout(sell2)

    /* 1.) 2 shout objects in db
    * - Sell 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
    * - Sell 1@10.01EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
    */
    assertEquals(2, Shout.list().size())

    Shout.list().each {shout ->
      assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
    }

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 10.0)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    def s2 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.01)
      eq('remainingQuantity', 1.0)
    }
    assertNotNull s2
    def tid2 = s2.transactionID
    def sid2 = s2.shoutID

    /* 2.) 1 TransactionLog entry:
    * - Quote: Ask: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
    */

    assertEquals(1, TransactionLog.list().size())
    TransactionLog tl1 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('ask', 10.0)
      eq('askSize', 10.0)
      eq('bidSize', 0.0)
      isNull('bid')
    }
    assertNotNull tl1
    assertEquals(tid1, tl1.transactionID)

    /* 3.) 2 Orderbook entries:
    * - ask1: 10 EUR, askSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
    * - ask1: 10EUR, askSize1: 10, ask2: 10.01 EUR, askSize2: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
    */
    assertEquals(2, Orderbook.list().size())

    Orderbook ob1 = Orderbook.withCriteria(uniqueResult: true) {
      eq('ask1', 10.0)
      eq('askSize1', 10.0)
      eq('bidSize1', 0.0)
      eq('askSize2', 0.0)
      isNull('ask2')
      isNull('bid1')
    }
    assertNotNull ob1
    assertEquals(tid1, ob1.transactionID)

    Orderbook ob2 = Orderbook.withCriteria(uniqueResult: true) {
      eq('ask1', 10.0)
      eq('askSize1', 10.0)
      eq('ask2', 10.01)
      eq('askSize2', 1.0)
      eq('bidSize1', 0.0)
      eq('askSize3', 0.0)
      isNull('ask3')
      isNull('bid1')
    }
    assertNotNull ob2
    assertEquals(tid2, ob2.transactionID)
  }

  /**
   * Init: New Buy 10@10EUR
   * Action: New Buy 1@9.99EUR
   * Validate:
   * 1.) 2 shout objects in db
   * - Buy 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Buy 1@9.99EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   *
   * 2.) 1 TransactionLog entry:
   * - Quote: Bid: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
   *
   * 3.) 2 Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - bid1: 10EUR, bidSize1: 10, bid2: 9.99 EUR, bidSize2: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewBuyOrderEntryCreatesNewOrderbookOnly() {
    //init
    buy1.limitPrice = 10.0
    buy1.remainingQuantity = 10.0
    auctionService.addShout(buy1)
    //action
    buy2.limitPrice = 9.99
    buy2.remainingQuantity = 1.0
    auctionService.addShout(buy2)

    /* 1.) 2 shout objects in db
    * - Buy 10@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
    * - Buy 1@9.99EUR, ModeReasonCode:1, transactionId: 2, , shoutId: 2
    */
    assertEquals(2, Shout.list().size())

    Shout.list().each {shout ->
      assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
    }

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 10.0)
      eq('remainingQuantity', 10.0)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    def s2 = Shout.withCriteria(uniqueResult: true) {
      eq('limitPrice', 9.99)
      eq('remainingQuantity', 1.0)
    }
    assertNotNull s2
    def tid2 = s2.transactionID
    def sid2 = s2.shoutID

    /* 2.) 1 TransactionLog entry:
    * - Quote: Bid: 10, AskSize: 10, Bid: null, BidSize: 0, transactionId: 1
    */
    assertEquals(1, TransactionLog.list().size())

    TransactionLog tl1 = TransactionLog.withCriteria(uniqueResult: true) {
      eq('bid', 10.0)
      eq('bidSize', 10.0)
      eq('askSize', 0.0)
      isNull('ask')
    }
    assertNotNull tl1
    assertEquals(tid1, tl1.transactionID)

    /* 3.) 2 Orderbook entries:
    * - bid1: 10 EUR, bidSize1: 10, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
    * - bid1: 10EUR, bidSize1: 10, bid2: 9.99 EUR, bidSize2: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
    */
    assertEquals(2, Orderbook.list().size())

    Orderbook ob1 = Orderbook.withCriteria(uniqueResult: true) {
      eq('bid1', 10.0)
      eq('bidSize1', 10.0)
      eq('askSize1', 0.0)
      eq('bidSize2', 0.0)
      isNull('bid2')
      isNull('ask1')
    }
    assertNotNull ob1
    assertEquals(tid1, ob1.transactionID)

    Orderbook ob2 = Orderbook.withCriteria(uniqueResult: true) {
      eq('bid1', 10.0)
      eq('bidSize1', 10.0)
      eq('bid2', 9.99)
      eq('bidSize2', 1.0)
      isNull('bid3')
      eq('bidSize3', 0.0)
      isNull('ask1')
      eq('askSize1', 0.0)
    }
    assertNotNull ob2
    assertEquals(tid2, ob2.transactionID)
  }

//----------------------------------------------------------------------------------------------------------------------

  /**
   * Init: Empty  db
   * Action: New Sell 50 (no limit)
   * Validate:
   * 1.) 2 shout objects in db
   * - Sell 50@10EUR, ModeReasonCode:1; shoutId: 1; transactionId: 1
   * - Sell 50@10EUR, DELETIONBYSYSTEM(6); shoutId: 1; transactionId: 1
   * 2.) No transactionLog Entries
   * 3.) No Orderbook entries
   */
  void testNewSellMarketOrderMeetsEmptyOrderBook() {
    //action
    msell.remainingQuantity = 50.0
    auctionService.addShout(msell)
    /* 1.) 2 shout objects in db
    * - Sell 50@10EUR, ModeReasonCode:1; shoutId: 1; transactionId: 1
    * - Sell 50@10EUR, DELETIONBYSYSTEM(6); shoutId: 1; transactionId: 1
    */
    assertEquals(2, Shout.list().size())

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('remainingQuantity', 50.0)
      eq('modReasonCode', ModReasonCode.INSERT)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    Shout s2 = Shout.withCriteria(uniqueResult: true) {
      eq('remainingQuantity', 50.0)
      eq('modReasonCode', ModReasonCode.DELETIONBYSYSTEM)
    }
    assertNotNull s2
    assertEquals(tid1, s2.transactionID)
    assertEquals(sid1, s2.shoutID)
    // 2.) No transactionLog Entries
    assertEquals(0, TransactionLog.list().size())
    // 3.) No Orderbook entries
    assertEquals(0, Orderbook.list().size())

  }

  /**
   * Init: Empty
   * Action: New Buy 50 (no limit)
   * Validate:
   * 1.) 2 shout objects in db
   * - Buy 50@10EUR, ModeReasonCode:1; shoutId:1. transactionId: 1
   * - Buy 50@10EUR, DELETIONBYSYSTEM(6); shoutId: 1, transactionId: 1
   * 2.) No transactionLog Entries
   * 3.) No Orderbook entries
   */
  void testNewBuyMarketOrderMeetsEmptyOrderBook() {
    //init + action
    mbuy.remainingQuantity = 50.0
    auctionService.addShout(mbuy)
   /* 1.) 2 shout objects in db
   * - Buy 50@10EUR, ModeReasonCode:1; shoutId:1. transactionId: 1
   * - Buy 50@10EUR, DELETIONBYSYSTEM(6); shoutId: 1, transactionId: 1
   */
    assertEquals(2, Shout.list().size())

    Shout s1 = Shout.withCriteria(uniqueResult: true) {
      eq('remainingQuantity', 50.0)
      eq('modReasonCode', ModReasonCode.INSERT)
    }
    assertNotNull s1
    def tid1 = s1.transactionID
    def sid1 = s1.shoutID

    Shout s2 = Shout.withCriteria(uniqueResult: true) {
      eq('remainingQuantity', 50.0)
      eq('modReasonCode', ModReasonCode.DELETIONBYSYSTEM)
    }
    assertNotNull s2
    assertEquals(tid1, s2.transactionID)
    assertEquals(sid1, s2.shoutID)
   //2.) No transactionLog Entries
    assertEquals (0, TransactionLog.list().size())
   //3.) No Orderbook entries
    assertEquals (0, Orderbook.list().size())
  }

  // ------------------------------------------------------------------------------------------------------------------

  /**
   * Init: New Buy 50@10EUR
   * Action: New Sell 100 (no limit)
   * Validate:
   * 1.) 5 shout objects in db
   * - Buy 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Sell 100@NULL, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   * - Sell 50@10EUR, ModeReasonCode: PARTIALEXECUTION(5), transactionId: 2, , shoutId: 2
   * - Buy 50@10EUR, EXECUTION(4), transactionId: 2, , shoutId: 1
   * - Sell 50 (no limit); DELETIONBYSYSTEM(6), transactionId: 2, shoutId: 2
   *
   * 2.) 3 TransactionLog entries:
   * - Quote: Bid: 10, BidSize: 50, ask: null, askSize: 0, transactionId: 1
   * - Trade: prize: 10, volume: 50, transactionId: 2
   * - Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 50, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - //Trade nicht im Orderbuch!!
   * - bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewSellMarketOrderEmptiesOrderBook() {
    //init
    buy1.remainingQuantity = 50.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)

    //action
    msell.remainingQuantity = 100.0
    auctionService.addShout(msell)

    //* 1.) 5 shout objects in db
    assertEquals(5, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@NULL, ModeReasonCode:1, transactionId: 2, , shoutId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Sell 50@10EUR, ModeReasonCode: PARTIALEXECUTION(5), transactionId: 2, , shoutId: 2
        case 3:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 50@10EUR, EXECUTION(4), transactionId: 2, , shoutId: 1
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 50 (no limit); DELETIONBYSYSTEM(6), transactionId: 2, shoutId: 2
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.DELETIONBYSYSTEM, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 3 TransactionLog entries:
    assertEquals(3, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 50, ask: null, askSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(50, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: prize: 10, volume: 50, transactionId: 2
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }
    
   // 3.) 2 Orderbook entries:
   assertEquals(2, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 50, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(50, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 2:
          assertEquals(msell.transactionID, orderbook.transactionID)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: New Sell 50@10EUR
   * Action: New Buy 100 (no limit)
   * Validate:
   * 1.) 5 shout objects in db
   * - Sell 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Buy 100@NULL, ModeReasonCode:1, transactionId: 2, , shoutId: 2
   * - Buy 50@10EUR, ModeReasonCode: PARTIALEXECUTION(5), transactionId: 2, , shoutId: 2
   * - Sell 50@10EUR, EXECUTION(4), transactionId: 2, , shoutId: 1
   * - Buy 50 (no limit); DELETIONBYSYSTEM(6), transactionId: 2, shoutId: 2
   *
   * 2.) 3 TransactionLog entries:
   * - Quote: Ask: 10, AskSize: 50, bid: null, bidSize: 0, transactionId: 1
   * - Trade: price: 10, volume: 50, transactionId: 2
   * - Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - ask1: 10 EUR, askSize1: 50, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - //Trade nicht im Orderbuch!!
   * - bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testNewBuyMarketOrderEmptiesOrderBook() {
    //init
    sell1.remainingQuantity = 50.0
    sell1.limitPrice = 10.0
    auctionService.addShout(sell1)

    //action
    mbuy.remainingQuantity = 100.0
    auctionService.addShout(mbuy)

    //* 1.) 5 shout objects in db
    assertEquals(5, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Sell 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
        case 1:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 100@NULL, ModeReasonCode:1, transactionId: 2, , shoutId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Buy 50@10EUR, ModeReasonCode: PARTIALEXECUTION(5), transactionId: 2, , shoutId: 2
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Sell 50@10EUR, EXECUTION(4), transactionId: 2, , shoutId: 1
        case 4:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 50 (no limit); DELETIONBYSYSTEM(6), transactionId: 2, shoutId: 2
        case 5:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.DELETIONBYSYSTEM, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }


   //* 2.) 3 TransactionLog entries:
    assertEquals(3, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: 10, AskSize: 50, bid: null, bidSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.ask)
          assertEquals(50, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: prize: 10, volume: 50, transactionId: 2
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 2 Orderbook entries:
   assertEquals(2, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //ask1: 10 EUR, askSize1: 50, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
        case 1:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.ask1)
          assertEquals(50, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 2:
          assertEquals(mbuy.transactionID, orderbook.transactionID)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  //-------------------------------------------------------------------------------------------------------------------

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Sell 100@11EUR
   * Action: New Sell 50 (no limit)
   * Validate:
   * 1.) 5 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
   * - Sell 50, No Limit, ModeReasonCode 1, shoutId: 3, transactionId: 3
   * - Sell 50, 10EUR: ModReasonCode: 4; EXECUTION, shoutId: 3, transactionId: 3
   * - Buy 50@10EUR, ModReasonCode: 5; PARTIALEXECUTION, shoutId: 1, transactionId: 3
   *
   * 2.) 4 TransactionLog entries:
   * - Quote: Bid: 10, BidSize: 100, Ask: null, askSize: 0, transactionId: 1
   * - Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
   * - Trade: prize: 10, volume: 50, transactionId: 3
   * - Quote: Bid: 10, BidSize: 50, Ask: 11, AskSize: 100, transactionId: 3
   *
   * 3.) 3. Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - //Trade nicht im Orderbuch!!
   * - bid1: 10 EUR, bidSize1: 50, ask1: 11, askSize1: 100, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketSellOrderTriggersPartialExecutionOfBuyLimitOrder() {
    //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)

    //action
    msell.remainingQuantity = 50.0
    auctionService.addShout(msell)

    //* 1.) 5 shout objects in db
    assertEquals(5, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 50, No Limit, ModeReasonCode 1, shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Sell 50, 10EUR: ModReasonCode: 4; EXECUTION, shoutId: 3, transactionId: 3
        case 4:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 50@10EUR, ModReasonCode: 5; PARTIALEXECUTION, shoutId: 1, transactionId: 3
        case 5:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 4 TransactionLog entries:
    assertEquals(4, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, Ask: null, askSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: prize: 10, volume: 50, transactionId: 3
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 50, Ask: 11, AskSize: 100, transactionId: 3
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(50, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 3 Orderbook entries:
   assertEquals(3, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        //bid1: 10 EUR, bidSize1: 50, ask1: 11, askSize1: 100, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(msell.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(50, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Sell 100@11EUR
   * Action: New Sell 50 @ 9EUR
   * Validate:
   * 1.) 5 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
   * - Sell 50, 9 EUR, ModeReasonCode 1; shoutId: 3, transactionId: 3
   * - Sell 50, 10 EUR: ModReasonCode: 4; EXECUTION; shoutId: 3, transactionId: 3
   * - Buy 50@10EUR, ModReasonCode: 5; PARTIALEXECUTION; shoutId: 1, transactionId: 3
   *
   * 2.) 4 TransactionLog entries:
   * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 1
   * - Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
   * - Trade: prize: 10, volume: 50, transactionId: 3
   * - Quote: Bid: 10, BidSize: 50, Ask: 11, AskSize: 100, transactionId: 3
   *
   * 3.) 3. Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - //Trade nicht im Orderbuch!!
   * - bid1: 10 EUR, bidSize1: 50, ask1: 11, askSize1: 100, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketableLimitSellOrderTriggersPartialExecutionOfBuyLimitOrder() {
    //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)

    //action
    sell2.remainingQuantity = 50.0
    sell2.limitPrice = 9.0
    auctionService.addShout(sell2)

    //* 1.) 5 shout objects in db
    assertEquals(5, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 50, 9 EUR, ModeReasonCode 1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell2.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //Sell 50, 10 EUR: ModReasonCode: 4; EXECUTION; shoutId: 3, transactionId: 3
        case 4:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell2.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //Buy 50@10EUR, ModReasonCode: 5; PARTIALEXECUTION; shoutId: 1, transactionId: 3
        case 5:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell2.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 4 TransactionLog entries:
    assertEquals(4, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: prize: 10, volume: 50, transactionId: 3
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(sell2.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 50, Ask: 11, AskSize: 100, transactionId: 3
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(50, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertEquals(sell2.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 3 Orderbook entries:
   assertEquals(3, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        //bid1: 10 EUR, bidSize1: 50, ask1: 11, askSize1: 100, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(sell2.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(50, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Sell 100@11EUR
   * Action: New Buy 50 (no limit)
   * Validate:
   * 1.) 5 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
   * - Buy 50, No Limit, ModeReasonCode 1, shoutId: 3, transactionId: 3
   * - Buy 50, 11EUR: ModReasonCode: 4; EXECUTION shoutId: 3, transactionId: 3
   * - Sell 50@11EUR, ModReasonCode: 5; PARTIALEXECUTION shoutId: 2, transactionId: 3
   *
   * 2.) 4 TransactionLog entries:
   * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 1
   * - Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
   * - Trade: prize: 11, volume: 50, transactionId: 3
   * - Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 50, transactionId: 3
   *
   * 3.) 3. Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - //Trade nicht im Orderbuch!!
   * - bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 50, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketBuyOrderTriggersPartialExecutionOfSellLimitOrder() {
   //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)

    //action
    mbuy.remainingQuantity = 50.0
    auctionService.addShout(mbuy)

    //* 1.) 5 shout objects in db
    assertEquals(5, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 50, No Limit, ModeReasonCode 1, shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Buy 50, 11EUR: ModReasonCode: 4; EXECUTION shoutId: 3, transactionId: 3
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Sell 50@11EUR, ModReasonCode: 5; PARTIALEXECUTION shoutId: 2, transactionId: 3
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 4 TransactionLog entries:
    assertEquals(4, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, Ask: null, askSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: prize: 11, volume: 50, transactionId: 3
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(11, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 50, transactionId: 3
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(50, translog.askSize)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 3 Orderbook entries:
   assertEquals(3, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        //bid1: 10 EUR, bidSize1: 50, ask1: 11, askSize1: 100, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(mbuy.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(50, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Sell 100@11EUR
   * Action: New Buy 50 @ 12EUR
   * Validate:
   * 1.) 5 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
   * - Buy 50, 12 EUR, ModeReasonCode 1, shoutId: 3, transactionId: 3
   * - Buy 50, 11 EUR: ModReasonCode: 4; EXECUTION, shoutId: 3 transactionId: 3
   * - Sell 50@11EUR, ModReasonCode: 5; PARTIALEXECUTION, shoutId: 2, transactionId: 3
   *
   * 2.) 4 TransaltionLog entries:
   * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0, transactionId: 1
   * - Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
   * - Trade: prize: 11, volume: 50, transactionId: 3
   * - Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 50, transactionId: 3
   *
   * 3.) 3. Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - //Trade nicht im Orderbuch!!
   * - bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 50, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketableLimitBuyOrderTriggersPartialExecutionOfSellLimitOrder() {
   //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)

    //action
    buy2.remainingQuantity = 50.0
	buy2.limitPrice = 12.0
    auctionService.addShout(buy2)

    //* 1.) 5 shout objects in db
    assertEquals(5, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR; ModReasonCode: 1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 50, 12 EUR, ModeReasonCode 1, shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 50, 11 EUR: ModReasonCode: 4; EXECUTION, shoutId: 3 transactionId: 3
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Sell 50@11EUR, ModReasonCode: 5; PARTIALEXECUTION, shoutId: 2, transactionId: 3
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 4 TransactionLog entries:
    assertEquals(4, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, Ask: null, askSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 100, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: prize: 11, volume: 50, transactionId: 3
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(11, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(buy2.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, Ask: 11, AskSize: 50, transactionId: 3
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertEquals(11, translog.ask)
          assertEquals(50, translog.askSize)
          assertEquals(buy2.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 3 Orderbook entries:
   assertEquals(3, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, ask1: 11, askSize1: 100, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        //bid1: 10 EUR, bidSize1: 50, ask1: 11, askSize1: 100, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(11, orderbook.ask1)
          assertEquals(50, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Buy 200@9EUR, 3.) Buy 300@8EUR
   * Action: New SEll 350 no Limit
   * Validate:
   * 1.) 10 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1, shoutId: 1, TransactionId: 1
   * - Buy 200@9EUR, ModeReasonCode:1, shoutId: 2, TransactionId: 2
   * - Buy 300@8EUR, ModeReasonCode:1, shoutId: 3, TransactionId: 3
   * - Sell 350@ no limit; ModReasonCode: 1, shoutId: 4, TransactionId: 4
   * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
   * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4), shoutId: 1, TransactionId: 4
   * - Sell 200@ 9; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
   * - Buy 200@9EUR, ModeReasonCode:EXECUTION(4),, shoutId: 2 TransactionId: 4
   * - Sell 50@ 8; ModReasonCode: EXECUTION(4), shoutId: 4, TransactionId: 4
   * - Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5), shoutId: 3, TransactionId: 4
   *
   * 2.) 5 TradeLog Entries
   * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 TransactionId: 1
   * - //Zweite Order erzeugt keinen Quote
   * - //Zweite Order erzeugt keinen Quote
   * - Trade: Prize: 10, Volume 100, TransactionId: 4
   * - Trade: Prize: 9, Volume 200, TransactionId: 4
   * - Trade: Prize: 8, Volume 50, TransactionId: 4
   * - Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0, TransactionId: 4
   *
   * 3.) 4 Orderbook Entries
   * - bid1: 10 EUR, bidSize1: 100, TransactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, TransactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, bid3: 8EUR, bidSize3: 300, TransactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * // Trades nicht im Orderbuch!
   * - bid1: 8 EUR, bidSize1: 250, TransactionId: 4
   */
  void testNewMarketSellOrderClearsSeveralLimitBuyOrders() {
    //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 200.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 300.0
    buy3.limitPrice = 8.0
    auctionService.addShout(buy3)

    //action
    msell.remainingQuantity = 350.0
    auctionService.addShout(msell)

    //* 1.) 10 shout objects in db
    assertEquals(10, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1, shoutId: 1, TransactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 200@9EUR, ModeReasonCode:1, shoutId: 2, TransactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 300@8EUR, ModeReasonCode:1, shoutId: 3, TransactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Sell 350@ no limit; ModReasonCode: 1, shoutId: 4, TransactionId: 4
        case 4:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4), shoutId: 1, TransactionId: 4
        case 6:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 200@ 9; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
        case 7:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(9, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 200@9EUR, ModeReasonCode:EXECUTION(4),, shoutId: 2 TransactionId: 4
        case 8:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(9, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Sell 50@ 8; ModReasonCode: EXECUTION(4), shoutId: 4, TransactionId: 4
        case 9:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(8, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5), shoutId: 3, TransactionId: 4
        case 10:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(8, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 5 TransactionLog entries:
    assertEquals(5, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 TransactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100, TransactionId: 4
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 200, TransactionId: 4
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(200, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Trade: Prize: 8, Volume 50, TransactionId: 4
        case 4:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(8, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0, TransactionId: 4
        case 5:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(250, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 4 Orderbook entries:
   assertEquals(4, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100, TransactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, TransactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, bid3: 8EUR, bidSize3: 300, TransactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 8 EUR, bidSize1: 250, TransactionId: 4
        case 4:
          assertEquals(msell.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(250, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Buy 200@9EUR, 3.) Buy 300@8EUR
   * Action: New SEll 350 @ 7 EUR
   * Validate:
   * 1.) 10 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Buy 200@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
   * - Buy 300@8EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
   * - Sell 350@ 7 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
   * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
   * - Sell 200@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - Buy 200@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
   * - Sell 50@ 8; ModReasonCode: EXECUTION(4); shoutId: 4, transactionId: 4
   * - Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 3, transactionId: 4
   *
   * 2.) 5 TradeLot Entries
   * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 1
   * - //Zweite Order erzeugt keinen Quote
   * - //Zweite Order erzeugt keinen Quote
   * - Trade: Prize: 10, Volume 100 transactionId: 4
   * - Trade: Prize: 9, Volume 200 transactionId: 4
   * - Trade: Prize: 8, Volume 50 transactionId: 4
   * - Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0, transactionId: 4
   *
   * 3.) 4 Orderbook Entries
   * - bid1: 10 EUR, bidSize1: 100 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, bid3: 8EUR, bidSize3: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * // Trades nicht im Orderbuch!
   * - bid1: 8 EUR, bidSize1: 250 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketableLimitSellOrderClearsSeveralLimitBuyOrders1() {
   //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 200.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 300.0
    buy3.limitPrice = 8.0
    auctionService.addShout(buy3)

    //action
    sell1.remainingQuantity = 350.0
    sell1.limitPrice = 7.0
    auctionService.addShout(sell1)

    //* 1.) 10 shout objects in db
    assertEquals(10, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1, shoutId: 1, TransactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 200@9EUR, ModeReasonCode:1, shoutId: 2, TransactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 300@8EUR, ModeReasonCode:1, shoutId: 3, TransactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Sell 350@ 7 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4), shoutId: 1, TransactionId: 4
        case 6:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 200@ 9; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
        case 7:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(9, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 200@9EUR, ModeReasonCode:EXECUTION(4),, shoutId: 2 TransactionId: 4
        case 8:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(9, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Sell 50@ 8; ModReasonCode: EXECUTION(4), shoutId: 4, TransactionId: 4
        case 9:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(8, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5), shoutId: 3, TransactionId: 4
        case 10:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(8, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 5 TransactionLog entries:
    assertEquals(5, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 TransactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100, TransactionId: 4
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 200, TransactionId: 4
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(200, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 8, Volume 50, TransactionId: 4
        case 4:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(8, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0, TransactionId: 4
        case 5:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(250, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 4 Orderbook entries:
   assertEquals(4, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100, TransactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, TransactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, bid3: 8EUR, bidSize3: 300, TransactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 8 EUR, bidSize1: 250, TransactionId: 4
        case 4:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(250, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 100 @ 10EUR, 2.) Buy 200@9EUR, 3.) Buy 300@8EUR
   * Action: New SEll 350 @ 9 EUR
   * Validate:
   * 1.) 9 shout objects saved in db:
   * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Buy 200@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
   * - Buy 300@8EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
   * - Sell 350@ 9 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
   * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
   * - Sell 200@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - Buy 200@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
   *
   * 2.) 5 TradeLot Entries
   * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 1
   * - //Zweite Order erzeugt keinen Quote
   * - //Zweite Order erzeugt keinen Quote
   * - Trade: Prize: 10, Volume 100 transactionId: 4
   * - Trade: Prize: 9, Volume 200 transactionId: 4
   * - Quote: Bid: 8, BidSize: 300, Ask: 9, AskSize: 50 transactionId: 4
   *
   * 3.) 4 Orderbook Entries
   * - bid1: 10 EUR, bidSize1: 100 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, bid3: 8EUR, bidSize3: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * // Trades nicht im Orderbuch!
   * - bid1: 8 EUR, bidSize1: 300, ask1: 9EUR, askSize1: 50 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketableLimitSellOrderClearsSeveralLimitBuyOrders2() {
    //init
    buy1.remainingQuantity = 100.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 200.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 300.0
    buy3.limitPrice = 8.0
    auctionService.addShout(buy3)

    //action
    sell1.remainingQuantity = 350.0
    sell1.limitPrice = 9.0
    auctionService.addShout(sell1)

    //* 1.) 8 shout objects in db
    assertEquals(8, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 100@10EUR, ModeReasonCode:1, shoutId: 1, TransactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 200@9EUR, ModeReasonCode:1, shoutId: 2, TransactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 300@8EUR, ModeReasonCode:1, shoutId: 3, TransactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Sell 350@ no limit; ModReasonCode: 1, shoutId: 4, TransactionId: 4
        case 4:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5), shoutId: 4, TransactionId: 4
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
        case 6:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(10, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 200@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
        case 7:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(9, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 200@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
        case 8:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(9, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 4 TransactionLog entries:
    assertEquals(4, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100 transactionId: 4
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 200 transactionId: 4
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(200, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 8, BidSize: 300, Ask: 9, AskSize: 50 transactionId: 4
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertEquals(9, translog.ask)
          assertEquals(50, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 4 Orderbook entries:
   assertEquals(4, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 100 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9EUR, bidSize2: 200, bid3: 8EUR, bidSize3: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 8 EUR, bidSize1: 300, ask1: 9EUR, askSize1: 50 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(9, orderbook.ask1)
          assertEquals(50, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Sell 100 @ 11EUR, 2.) Sell 200@12EUR, 3.) Sell 300@13EUR
   * Action: New Buy 350 no Limit
   * Validate:
   * 1.) 10 shout objects saved in db:
   * - Sell 100@11EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 200@12EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
   * - Sell 300@13EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
   * - Buy 350@ no limit; ModReasonCode: 1; shoutId: 4, transactionId: 4
   * - Buy 100@ 11; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - Sell 100@11EUR, ModeReasonCode: EXECUTION(4);  shoutId: 1, transactionId: 4
   * - Buy 200@ 12; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - SEll 200@12EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
   * - Buy 50@ 13; ModReasonCode: EXECUTION(4); shoutId: 4, transactionId: 4
   * - Sell 50@13EUR, ModeReasonCode:PARTIALEXECUTION(5);  shoutId: 3, transactionId: 4
   *
   * 2.) 5 TradeLot Entries
   * - Quote: Ask: 11, AskSize: 100, bid: null, bidSize: 0 transactionId: 1
   * - //Zweite Order erzeugt keinen Quote
   * - //Zweite Order erzeugt keinen Quote
   * - Trade: Prize: 11, Volume 100 transactionId: 4
   * - Trade: Prize: 12, Volume 200 transactionId: 4
   * - Trade: Prize: 13, Volume 50 transactionId: 4
   * - Quote: Ask: 13, AskSize: 250, Bid: null, bidSize: 0 transactionId: 4
   *
   * 3.) 4 Orderbook Entries
   * - ask1: 11 EUR, bidSize1: 100 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200  transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, ask3: 13EUR, askSize3: 300  transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * // Trades nicht im Orderbuch!
   * - ask1: 13 EUR, askSize1: 250 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketBuyOrderClearsSeveralLimitSellOrders() {
    //init
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)
    sell2.remainingQuantity = 200.0
    sell2.limitPrice = 12.0
    auctionService.addShout(sell2)
    sell3.remainingQuantity = 300.0
    sell3.limitPrice = 13.0
    auctionService.addShout(sell3)

    //action
    mbuy.remainingQuantity = 350.0
    auctionService.addShout(mbuy)

    //* 1.) 13 shout objects in db
    assertEquals(10, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //SELL 100@11EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //SELL 200@12EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell2.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //SELL 300@13EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(13, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell3.transactionID, shout.transactionID)
          assertEquals(sell3.shoutID, shout.shoutID)
          break

        //BUY 350@ no limit; ModReasonCode: 1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //BUY  100@ 11; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
        case 5:
          assertEquals(BuySellIndicator.BUY , shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(11, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(11, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 200@ 12; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(12, shout.executionPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Sell 200@12EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(12, shout.executionPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //Buy 50@ 13; ModReasonCode: EXECUTION(5); shoutId: 3, transactionId: 4
        case 9:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(13, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(mbuy.shoutID, shout.shoutID)
          break

        //Sell 50@ 13; ModReasonCode: PARTIALEXECUTION(4); shoutId: 3, transactionId: 4
        case 10:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(13, shout.limitPrice)
          assertEquals(13, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(mbuy.transactionID, shout.transactionID)
          assertEquals(sell3.shoutID, shout.shoutID)
          break

         default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 5 TransactionLog entries:
    assertEquals(5, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: 11, AskSize: 100, bid: null, bidSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 11, Volume 100 transactionId: 4
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(11, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        //Trade: Prize: 12, Volume 200 transactionId: 4
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(12, translog.price)
          assertEquals(200, translog.quantity)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        //Trade: Prize: 13, Volume 50 transactionId: 4
        case 4:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(13, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        //Quote: Ask: 13, AskSize: 250, Bid: null, bidSize: 0 transactionId: 4
        case 5:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(13, translog.ask)
          assertEquals(250, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(mbuy.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 4 Orderbook entries:
   assertEquals(4, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //ask1: 11 EUR, bidSize1: 100 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200  transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell2.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertEquals(12, orderbook.ask2)
          assertEquals(200, orderbook.askSize2)
          assertNull(orderbook.ask3)
          assertEquals(0, orderbook.askSize3)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, ask3: 13EUR, askSize3: 300  transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(sell3.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertEquals(12, orderbook.ask2)
          assertEquals(200, orderbook.askSize2)
          assertEquals(13, orderbook.ask3)
          assertEquals(300, orderbook.askSize3)
          assertNull(orderbook.ask4)
          assertEquals(0, orderbook.askSize4)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 13 EUR, askSize1: 250 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 5:
          assertEquals(mbuy.transactionID, orderbook.transactionID)
          assertEquals(13, orderbook.ask1)
          assertEquals(250, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Sell 100 @ 11EUR, 2.) Sell 200@12EUR, 3.) Sell 300@13EUR
   * Action: New Buy 350 @ 14 EUR
   * Validate:
   * 1.) 10 shout objects saved in db:
   * - Sell 100@11EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 200@12EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
   * - Sell 300@13EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
   * - Buy 350@ 14 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
   * - Buy 100@ 11; ModReasonCode: PARTIALEXECUTION(5);  shoutId: 4, transactionId: 4
   * - Sell 100@11EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
   * - Buy 200@ 12; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - SEll 200@12EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
   * - Buy 50@ 13; ModReasonCode: EXECUTION(4); shoutId: 4, transactionId: 4
   * - Sell 50@13EUR, ModeReasonCode:PARTIALEXECUTION(5);  shoutId: 3, transactionId: 4
   *
   * 2.) 5 TradeLot Entries
   * - Quote: Ask: 11, AskSize: 100, bid: null, bidSize: 0, transactionId: 1
   * - //Zweite Order erzeugt keinen Quote
   * - //Zweite Order erzeugt keinen Quote
   * - Trade: Prize: 11, Volume 100, transactionId: 4
   * - Trade: Prize: 12, Volume 200, transactionId: 4
   * - Trade: Prize: 13, Volume 50, transactionId: 4
   * - Quote: Ask: 13, AskSize: 250, bid: null, bidSize: 0 transactionId: 4
   *
   * 3.) 4 Orderbook Entries
   * - ask1: 11 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, ask3: 13EUR, askSize3: 300, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * // Trades nicht im Orderbuch!
   * - ask1: 13 EUR, askSize1: 250, transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketableLimitBuyOrderClearsSeveralLimitSellOrders1() {
   //init
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)
    sell2.remainingQuantity = 200.0
    sell2.limitPrice = 12.0
    auctionService.addShout(sell2)
    sell3.remainingQuantity = 300.0
    sell3.limitPrice = 13.0
    auctionService.addShout(sell3)

    //action
    buy1.remainingQuantity = 350.0
    buy1.limitPrice = 14.0
    auctionService.addShout(buy1)

    //* 1.) 10 shout objects in db
    assertEquals(10, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Sell 100@11EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 200@12EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell2.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //Sell 300@13EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(13, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell3.transactionID, shout.transactionID)
          assertEquals(sell3.shoutID, shout.shoutID)
          break

        //Buy 350@ 14 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertEquals(14, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 100@ 11; ModReasonCode: PARTIALEXECUTION(5);  shoutId: 4, transactionId: 4
        case 5:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(14, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 200@ 12; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(14, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(12, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //SEll 200@12EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(12, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //Buy 50@ 13; ModReasonCode: EXECUTION(4); shoutId: 4, transactionId: 4
        case 9:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(14, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(13, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 50@13EUR, ModeReasonCode:PARTIALEXECUTION(5);  shoutId: 3, transactionId: 4
        case 10:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(13, shout.limitPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(13, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)                                                                              
          assertEquals(sell3.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 5 TransactionLog entries:
    assertEquals(5, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: 11, AskSize: 100, bid: null, bidSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 11, Volume 100, transactionId: 4
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(11, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 12, Volume 200, transactionId: 4
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(12, translog.price)
          assertEquals(200, translog.quantity)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 13, Volume 50, transactionId: 4
        case 4:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(13, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Ask: 13, AskSize: 250, bid: null, bidSize: 0 transactionId: 4
        case 5:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(13, translog.ask)
          assertEquals(250, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 4 Orderbook entries:
   assertEquals(4, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //ask1: 11 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell2.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertEquals(12, orderbook.ask2)
          assertEquals(200, orderbook.askSize2)
          assertNull(orderbook.ask3)
          assertEquals(0, orderbook.askSize3)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, ask3: 13EUR, askSize3: 300, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(sell3.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertEquals(12, orderbook.ask2)
          assertEquals(200, orderbook.askSize2)
          assertEquals(13, orderbook.ask3)
          assertEquals(300, orderbook.askSize3)
          assertNull(orderbook.ask4)
          assertEquals(0, orderbook.askSize4)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 13 EUR, askSize1: 250, transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(13, orderbook.ask1)
          assertEquals(250, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  /**
   * Init: Create and save (in order):  1.) Sell 100 @ 11EUR, 2.) Sell 200@12EUR, 3.) Sell 300@13EUR
   * Action: New Buy 350 @ 12 EUR
   * Validate:
   * 1.) 8 shout objects saved in db:
   * - Sell 100@11EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
   * - Sell 200@12EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
   * - Sell 300@13EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
   * - Buy 350@ 12 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
   * - Buy 100@ 11; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - Sell 100@11EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
   * - Buy 200@ 12; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
   * - SEll 200@12EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
   *
   * 2.) 4 TradeLot Entries
   * - Quote: Ask: 11, AskSize: 100, bid: null, bidSize: 0 transactionId: 1
   * - //Zweite Order erzeugt keinen Quote
   * - //Zweite Order erzeugt keinen Quote
   * - Trade: Prize: 11, Volume 100, transactionId: 4
   * - Trade: Prize: 12, Volume 200, transactionId: 4
   * - Quote: Bid: 12, BidSize: 50, Ask: 13, AskSize: 300, transactionId: 4
   *
   * 3.) 4 Orderbook Entries
   * - ask1: 11 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * - ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, ask3: 13EUR, askSize3: 300, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
   * // Trades nicht im Orderbuch!
   * - bid1: 12, bidSize1: 50, ask1: 13 EUR, askSize1: 300, transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
   */
  void testNewMarketableLimitBuyOrderClearsSeveralLimitSellOrders2() {
   //init
    sell1.remainingQuantity = 100.0
    sell1.limitPrice = 11.0
    auctionService.addShout(sell1)
    sell2.remainingQuantity = 200.0
    sell2.limitPrice = 12.0
    auctionService.addShout(sell2)
    sell3.remainingQuantity = 300.0
    sell3.limitPrice = 13.0
    auctionService.addShout(sell3)

    //action
    buy1.remainingQuantity = 350.0
    buy1.limitPrice = 12.0
    auctionService.addShout(buy1)

    //* 1.) 8 shout objects in db
    assertEquals(8, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Sell 100@11EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 200@12EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell2.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        //Sell 300@13EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(13, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell3.transactionID, shout.transactionID)
          assertEquals(sell3.shoutID, shout.shoutID)
          break

        //Buy 350@ 12 EUR; ModReasonCode: 1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 100@ 11; ModReasonCode: PARTIALEXECUTION(5);  shoutId: 4, transactionId: 4
        case 5:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Sell 100@11EUR, ModeReasonCode: EXECUTION(4); shoutId: 1, transactionId: 4
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(11, shout.limitPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(11, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 200@ 12; ModReasonCode: PARTIALEXECUTION(5); shoutId: 4, transactionId: 4
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(12, shout.executionPrice)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //SEll 200@12EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 4
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(12, shout.limitPrice)
          assertEquals(200, shout.executionQuantity)
          assertEquals(12, shout.executionPrice)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(sell2.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 4 TransactionLog entries:
    assertEquals(4, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: 11, AskSize: 100, bid: null, bidSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(11, translog.ask)
          assertEquals(100, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 11, Volume 100, transactionId: 4
        case 2:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(11, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 12, Volume 200, transactionId: 4
        case 3:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(12, translog.price)
          assertEquals(200, translog.quantity)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 12, BidSize: 50, Ask: 13, AskSize: 300, transactionId: 4
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(13, translog.ask)
          assertEquals(300, translog.askSize)
          assertEquals(12, translog.bid)
          assertEquals(50, translog.bidSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 4 Orderbook entries:
   assertEquals(4, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //ask1: 11 EUR, bidSize1: 100, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(sell2.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertEquals(12, orderbook.ask2)
          assertEquals(200, orderbook.askSize2)
          assertNull(orderbook.ask3)
          assertEquals(0, orderbook.askSize3)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //ask1: 11 EUR, bidSize1: 100, ask2: 12EUR, askSize2: 200, ask3: 13EUR, askSize3: 300, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(sell3.transactionID, orderbook.transactionID)
          assertEquals(11, orderbook.ask1)
          assertEquals(100, orderbook.askSize1)
          assertEquals(12, orderbook.ask2)
          assertEquals(200, orderbook.askSize2)
          assertEquals(13, orderbook.ask3)
          assertEquals(300, orderbook.askSize3)
          assertNull(orderbook.ask4)
          assertEquals(0, orderbook.askSize4)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //bid1: 12, bidSize1: 50, ask1: 13 EUR, askSize1: 300, transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(13, orderbook.ask1)
          assertEquals(300, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          assertEquals(12, orderbook.bid1)
          assertEquals(50, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

  // --------------------------------------------------------------------------

/**
 * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR, 3.) Buy 50@9EUR, 4.) Buy 100@10EUR
 * Action: New SEll 350 no Limit
 * Validate:
 * 1.) 13 shout objects saved in db:
 * - Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
 * - Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
 * - Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
 * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
 * - Sell 350@ no limit; ModReasonCode: 1; shoutId: 5, transactionId: 5
 * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
 * - Sell 150@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 150@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 has to be ececuted first!
 * - Sell 50@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - BUY 50@ 9; ModReasonCode: EXECUTION(4); shoutId: 3, transactionId: 5
 * - Sell 50@ 8; ModReasonCode: EXECUTION(4); shoutId: 5, transactionId: 5
 * - Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 1, transactionId: 5
 *
 * 2.) 8 TradeLog Entries
 * - Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0 transactionId: 1
 * - Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0 transactionId: 2
 * - Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0 transactionId: 3
 * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 4
 * - Trade: Prize: 10, Volume 100 transactionId: 5
 * - Trade: Prize: 9, Volume 150 transactionId: 5
 * - Trade: Prize: 9, Volume 50 transactionId: 5
 * - Trade: Prize: 8, Volume 50 transactionId: 5
 * - Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0 transactionId: 5
 *
 * 3.) 5 Orderbook Entries
 * - bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * // Trades nicht im Orderbuch!
 * - bid1: 8 EUR, bidSize1: 250 transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
 */
  void testNewMarketSellOrderClearsSeveralLimitBuyOrdersWithTimePriority() {
    //init
    buy1.remainingQuantity = 300.0
    buy1.limitPrice = 8.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 150.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 50.0
    buy3.limitPrice = 9.0
    auctionService.addShout(buy3)
    buy4.remainingQuantity = 100.0
    buy4.limitPrice = 10.0
    auctionService.addShout(buy4)

    //action
    msell.remainingQuantity = 350.0
    auctionService.addShout(msell)

    //* 1.) 13 shout objects in db
    assertEquals(13, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy4.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 350@ no limit; ModReasonCode: 1; shoutId: 5, transactionId: 5
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 150@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(150, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 has to be ececuted first!
        case 9:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(150, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Sell 50@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 10:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //BUY 50@ 9; ModReasonCode: EXECUTION(4); shoutId: 3, transactionId: 5
        case 11:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Sell 50@ 8; ModReasonCode: EXECUTION(4); shoutId: 5, transactionId: 5
        case 12:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(8, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 1, transactionId: 5
        case 13:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(8, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 8 TransactionLog entries:
    assertEquals(9, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0 transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(150, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy2.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0 transactionId: 3
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(200, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy3.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 4
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy4.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100 transactionId: 5
        case 5:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 150 transactionId: 5
        case 6:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(150, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 50 transactionId: 5
        case 7:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Trade: Prize: 8, Volume 50 transactionId: 5
        case 8:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(8, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0 transactionId: 5
        case 9:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(250, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 5 Orderbook entries:
   assertEquals(5, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(200, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(buy4.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 8 EUR, bidSize1: 250 transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 5:
          assertEquals(msell.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(250, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

/**
 * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR, 3.) Buy 50@9EUR, 4.) Buy 100@10EUR
 * Action: New SEll 200 no Limit
 * Validate:
 * 1.) 9 shout objects saved in db:
 * - Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
 * - Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
 * - Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
 * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
 * - Sell 200@ no limit; ModReasonCode: 1; shoutId: 5, transactionId: 5
 * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
 * - Sell 100@ 9; ModReasonCode: EXECUTION(4); shoutId: 5, transactionId: 5
 * - Buy 100@9EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 order has to be ececuted first!
 *
 * 2.) 8 TradeLog Entries
 * - Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0, transactionId: 1
 * - Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0, transactionId: 2
 * - Quote: Bid: 9, BidSize: 200,  ask: null, askSize: 0, transactionId: 3
 * - Quote: Bid: 10, BidSize: 100,  ask: null, askSize: 0, transactionId: 4
 * - Trade: Prize: 10, Volume 100, transactionId: 5
 * - Trade: Prize: 9, Volume 100, transactionId: 5
 * - Quote: Bid: 9, BidSize: 100,  ask: null, askSize: 0, transactionId: 5
 *
 * 3.) 5 Orderbook Entries
 * - bid1: 8 EUR, bidSize1: 300, transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300, transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300, transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300, transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * // Trades nicht im Orderbuch!
 * - bid1: 9 EUR, bidSize1: 100, bid2: 8 EUR, bidSize2: 300, transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
 */
  void testNewMarketSellOrderClearsSeveralLimitBuyOrdersPartiallyWithTimePriority() {
    //init
    buy1.remainingQuantity = 300.0
    buy1.limitPrice = 8.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 150.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 50.0
    buy3.limitPrice = 9.0
    auctionService.addShout(buy3)
    buy4.remainingQuantity = 100.0
    buy4.limitPrice = 10.0
    auctionService.addShout(buy4)

    //action
    msell.remainingQuantity = 200.0
    auctionService.addShout(msell)

    //* 1.) 9 shout objects in db
    assertEquals(9, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy4.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 200@ no limit; ModReasonCode: 1; shoutId: 5, transactionId: 5
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(200, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 100@ 9; ModReasonCode: EXECUTION(4); shoutId: 5, transactionId: 5
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertNull(shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(msell.shoutID, shout.shoutID)
          break

        //Buy 100@9EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 order has to be ececuted first!
        case 9:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(msell.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 8 TransactionLog entries:
    assertEquals(7, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0 transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(150, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy2.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0 transactionId: 3
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(200, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy3.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 4
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy4.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100 transactionId: 5
        case 5:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 100 transactionId: 5
        case 6:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 100,  ask: null, askSize: 0, transactionId: 5
        case 7:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(msell.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 5 Orderbook entries:
   assertEquals(5, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(200, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(buy4.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 100, bid2: 8 EUR, bidSize2: 300, transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 5:
          assertEquals(msell.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

/**
 * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR, 3.) Buy 50@9EUR, 4.) Buy 100@10EUR
 * Action: New SEll 350 @ 7 EUR
 * Validate:
 * 1.) 13 shout objects saved in db:
 * - Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
 * - Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
 * - Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
 * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
 * - Sell 350@ 7 EUR; ModReasonCode: 1; shoutId: 5, transactionId: 5
 * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
 * - Sell 150@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 150@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 has to be ececuted first!
 * - Sell 50@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 3, transactionId: 5
 * - BUY 50@ 9; ModReasonCode: EXECUTION(4); shoutId: 3, transactionId: 5
 * - Sell 50@ 8; ModReasonCode: EXECUTION(4); shoutId: 5, transactionId: 5
 * - Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 1, transactionId: 5
 *
 * 2.) 8 TradeLog Entries
 * - Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0, transactionId: 1
 * - Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0, transactionId: 2
 * - Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0, transactionId: 3
 * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0, transactionId: 4
 * - Trade: Prize: 10, Volume 100 transactionId: 5
 * - Trade: Prize: 9, Volume 150 transactionId: 5
 * - Trade: Prize: 9, Volume 50 transactionId: 5
 * - Trade: Prize: 8, Volume 50 transactionId: 5
 * - Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0, transactionId: 5
 *
 * 3.) 5 Orderbook Entries
 * - bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * // Trades nicht im Orderbuch!
 * - bid1: 8 EUR, bidSize1: 250 transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
 */
  void testNewMarketSellLimitOrderClearsSeveralLimitBuyOrdersWithTimePriority1() {
   //init
    buy1.remainingQuantity = 300.0
    buy1.limitPrice = 8.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 150.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 50.0
    buy3.limitPrice = 9.0
    auctionService.addShout(buy3)
    buy4.remainingQuantity = 100.0
    buy4.limitPrice = 10.0
    auctionService.addShout(buy4)

    //action
    sell1.remainingQuantity = 350.0
    sell1.limitPrice = 7.0
    auctionService.addShout(sell1)

    //* 1.) 13 shout objects in db
    assertEquals(13, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy4.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 350@ 7 EUR; ModReasonCode: 1; shoutId: 5, transactionId: 5
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 150@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(150, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 has to be ececuted first!
        case 9:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(150, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Sell 50@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 3, transactionId: 5
        case 10:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //BUY 50@ 9; ModReasonCode: EXECUTION(4); shoutId: 3, transactionId: 5
        case 11:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Sell 50@ 8; ModReasonCode: EXECUTION(4); shoutId: 5, transactionId: 5
        case 12:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(7, shout.limitPrice)
          assertEquals(8, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 50@8EUR, ModeReasonCode:PARTIALEXECUTION(5); shoutId: 1, transactionId: 5
        case 13:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(8, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 8 TransactionLog entries:
    assertEquals(9, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0 transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(150, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy2.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0 transactionId: 3
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(200, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy3.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 4
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy4.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100 transactionId: 5
        case 5:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 150 transactionId: 5
        case 6:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(150, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 50 transactionId: 5
        case 7:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 8, Volume 50 transactionId: 5
        case 8:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(8, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 8, BidSize: 250, ask: null, askSize: 0 transactionId: 5
        case 9:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(250, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 5 Orderbook entries:
   assertEquals(5, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(200, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(buy4.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 8 EUR, bidSize1: 250 transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 5:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(250, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

/**
 * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR, 3.) Buy 50@9EUR, 4.) Buy 100@10EUR
 * Action: New SEll 350 @ 9 EUR
 * Validate:
 * 1.) 11 shout objects saved in db:
 * - Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
 * - Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
 * - Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
 * - Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
 * - Sell 350@ 9 EUR; ModReasonCode: 1; shoutId: 5, transactionId: 5
 * - Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
 * - Sell 150@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - Buy 150@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 has to be ececuted first!
 * - Sell 50@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
 * - BUY 50@ 9; ModReasonCode: EXECUTION(4); shoutId: 3, transactionId: 5
 *
 * 2.) 7 TradeLog Entries
 * - Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0, transactionId: 1
 * - Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0, transactionId: 2
 * - Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0, transactionId: 3
 * - Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0, transactionId: 4
 * - Trade: Prize: 10, Volume 100 transactionId: 5
 * - Trade: Prize: 9, Volume 150 transactionId: 5
 * - Trade: Prize: 9, Volume 50 transactionId: 5
 * - Quote: Bid: 8, BidSize: 300, Ask: 9, AskSize: 50 transactionId: 5
 *
 * 3.) 5 Orderbook Entries
 * - bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * - bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
 * // Trades nicht im Orderbuch!
 * - bid1: 8 EUR, bidSize1: 300, ask1: 9, askSize1: 50 transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
 */
  void testNewMarketSellLimitOrderClearsSeveralLimitBuyOrdersWithTimePriority2() {
   //init
    buy1.remainingQuantity = 300.0
    buy1.limitPrice = 8.0
    auctionService.addShout(buy1)
    buy2.remainingQuantity = 150.0
    buy2.limitPrice = 9.0
    auctionService.addShout(buy2)
    buy3.remainingQuantity = 50.0
    buy3.limitPrice = 9.0
    auctionService.addShout(buy3)
    buy4.remainingQuantity = 100.0
    buy4.limitPrice = 10.0
    auctionService.addShout(buy4)

    //action
    sell1.remainingQuantity = 350.0
    sell1.limitPrice = 9.0
    auctionService.addShout(sell1)

    //* 1.) 11 shout objects in db
    assertEquals(11, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 300@8EUR, ModeReasonCode:1; shoutId: 1, transactionId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:1; shoutId: 2, transactionId: 2
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Buy 50@9EUR, ModeReasonCode:1; shoutId: 3, transactionId: 3
        case 3:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy3.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode:1; shoutId: 4, transactionId: 4
        case 4:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy4.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 350@ 7 EUR; ModReasonCode: 1; shoutId: 5, transactionId: 5
        case 5:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(350, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Sell 100@ 10; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 6:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(250, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 100@10EUR, ModeReasonCode: EXECUTION(4); shoutId: 4, transactionId: 5
        case 7:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(10, shout.executionPrice)
          assertEquals(100, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy4.shoutID, shout.shoutID)
          break

        //Sell 150@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 8:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(100, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(150, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:EXECUTION(4); shoutId: 2, transactionId: 5  !!! Note Time-Precedence: 150 has to be ececuted first!
        case 9:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(150, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //Sell 50@ 9; ModReasonCode: PARTIALEXECUTION(5); shoutId: 5, transactionId: 5
        case 10:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.PARTIALEXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //BUY 50@ 9; ModReasonCode: EXECUTION(4); shoutId: 3, transactionId: 5
        case 11:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(0, shout.remainingQuantity)
          assertEquals(9, shout.limitPrice)
          assertEquals(9, shout.executionPrice)
          assertEquals(50, shout.executionQuantity)
          assertEquals(ModReasonCode.EXECUTION, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(buy3.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }

   //* 2.) 7 TransactionLog entries:
    assertEquals(8, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Bid: 8, BidSize: 300, ask: null, askSize: 0 transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 150, ask: null, askSize: 0 transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(150, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy2.transactionID, translog.transactionID)
          break

        //Quote: Bid: 9, BidSize: 200, ask: null, askSize: 0 transactionId: 3
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(9, translog.bid)
          assertEquals(200, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy3.transactionID, translog.transactionID)
          break

        //Quote: Bid: 10, BidSize: 100, ask: null, askSize: 0 transactionId: 4
        case 4:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(10, translog.bid)
          assertEquals(100, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(buy4.transactionID, translog.transactionID)
          break

        //Trade: Prize: 10, Volume 100 transactionId: 5
        case 5:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(10, translog.price)
          assertEquals(100, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 150 transactionId: 5
        case 6:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(150, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Trade: Prize: 9, Volume 50 transactionId: 5
        case 7:
          assertEquals(TransactionType.TRADE, translog.type)
          assertEquals(9, translog.price)
          assertEquals(50, translog.quantity)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Quote: Bid: 8, BidSize: 300, Ask: 9, AskSize: 50 transactionId: 5
        case 8:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertEquals(9, translog.ask)
          assertEquals(50, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 5 Orderbook entries:
   assertEquals(5, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 8 EUR, bidSize1: 300 transactionId: 1, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300 transactionId: 2, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 2:
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 9 EUR, bidSize1: 200, bid2: 8 EUR, bidSize2: 300 transactionId: 3, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 3:
          assertEquals(buy3.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(200, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 10 EUR, bidSize1: 100, bid2: 9 EUR, bidSize2: 200, bid3: 8, bidSize3: 300 transactionId: 4, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 4:
          assertEquals(buy4.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(100, orderbook.bidSize1)
          assertEquals(9, orderbook.bid2)
          assertEquals(200, orderbook.bidSize2)
          assertEquals(8, orderbook.bid3)
          assertEquals(300, orderbook.bidSize3)
          assertNull(orderbook.bid4)
          assertEquals(0, orderbook.bidSize4)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: 8 EUR, bidSize1: 300, ask1: 9, askSize1: 50 transactionId: 5, all other ask* and bid*==null, all other bidSize* and askSize*=0
        case 5:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertEquals(9, orderbook.ask1)
          assertEquals(50, orderbook.askSize1)
          assertNull(orderbook.ask2)
          assertEquals(0, orderbook.askSize2)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

    /**
   * Init: New Buy 50@10EUR
   * Action: Del Buy Order
   * Validate:
   * 1.) 2 shout objects in db
   * - Buy 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Del, ModeReasonCode:del by user, transactionId: 2, , shoutId: 1
   *
   * 2.) 2 TransactionLog entries:
   * - Quote: Ask: 10, AskSize: 50, bid: null, bidSize: 0, transactionId: 1
   * - Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - bid1: 10 EUR, bidSize1: 50, all other bid* and ask*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testDelBuyLimitOrderToEmptyOrderbook() {
    //init
    buy1.remainingQuantity = 50.0
    buy1.limitPrice = 10.0
    auctionService.addShout(buy1)

    //action
    auctionService.deleteShout(buy1)

    //* 1.) 5 shout objects in db
    assertEquals(2, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
        case 1:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Del, ModeReasonCode:del by user, transactionId: 2, , shoutId: 1
        case 2:
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.DELETIONBYUSER, shout.modReasonCode)
          assertNotSame(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, buy1.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }


   //* 2.) 3 TransactionLog entries:
    assertEquals(2, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: 10, AskSize: 50, bid: null, bidSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(10, translog.bid)
          assertEquals(50, translog.bidSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertNotSame(buy1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 2 Orderbook entries:
   assertEquals(2, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 50, all other bid* and ask*==null, all other bidSize* and askSize*=0, transactionId: 1
        case 1:
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.bid1)
          assertEquals(50, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        //bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 2:
          assertNotSame(buy1.transactionID, orderbook.transactionID)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
   }
  }

   /**
   * Init: New Sell 50@10EUR
   * Action: Del Sell Order
   * Validate:
   * 1.) 2 shout objects in db
   * - Sell 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
   * - Del, ModeReasonCode:del by user, transactionId: 2, , shoutId: 1
   *
   * 2.) 2 TransactionLog entries:
   * - Quote: Ask: null, AskSize: 0, bid: 10, bidSize: 50, transactionId: 1
   * - Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
   *
   * 3.) 2 Orderbook entries:
   * - ask1: 10 EUR, askSize1: 50, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 1
   * - bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
   */
  void testDelSellLimitOrderToEmptyOrderbook() {
   //init
    sell1.remainingQuantity = 50.0
    sell1.limitPrice = 10.0
    auctionService.addShout(sell1)

    //action
    auctionService.deleteShout(sell1)

    //* 1.) 5 shout objects in db
    assertEquals(2, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 50@10EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
        case 1:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        //Del, ModeReasonCode:del by user, transactionId: 2, , shoutId: 1
        case 2:
          assertEquals(BuySellIndicator.SELL, shout.buySellIndicator)
          assertEquals(50, shout.remainingQuantity)
          assertEquals(10, shout.limitPrice)
          assertEquals(ModReasonCode.DELETIONBYUSER, shout.modReasonCode)
          assertNotSame(sell1.transactionID, shout.transactionID)
          assertEquals(sell1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }


   //* 2.) 3 TransactionLog entries:
    assertEquals(2, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: 10, AskSize: 50, bid: null, bidSize: 0, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertEquals(10, translog.ask)
          assertEquals(50, translog.askSize)
          assertEquals(sell1.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: null, AskSize: 0, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertNull(translog.bid)
          assertEquals(0, translog.bidSize)
          assertNotSame(sell1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 2 Orderbook entries:
   assertEquals(2, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 10 EUR, bidSize1: 50, all other bid* and ask*==null, all other bidSize* and askSize*=0, transactionId: 1
        case 1:
          assertEquals(sell1.transactionID, orderbook.transactionID)
          assertEquals(10, orderbook.ask1)
          assertEquals(50, orderbook.askSize1)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          break

        //bid1: null EUR, bidSize1: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 2:
          assertNotSame(sell1.transactionID, orderbook.transactionID)
          assertNull(orderbook.bid1)
          assertEquals(0, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
    }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR
   * Action: Delete Order 2.)
   */

  void testDelSecondOfTwoBuyOrders() {
   //init
    buy1.remainingQuantity = 300
    buy1.limitPrice = 8.00
    auctionService.addShout(buy1)

    buy2.remainingQuantity = 150
    buy2.limitPrice = 9.00
    auctionService.addShout(buy2)

    //action
    auctionService.deleteShout(buy2)

    //* 1.) 3 shout objects in db
    assertEquals(3, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 300@8EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
        case 1:
          println(shout.id +" "+ shout.remainingQuantity + " " + shout.limitPrice + " "+ shout.modReasonCode)
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:1, transactionId: 2, shoutId: 2
        case 2:
          println(shout.id +" "+ shout.remainingQuantity + " " + shout.limitPrice + " "+ shout.modReasonCode)
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9.00, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //DEL 150@9EUR, ModeReasonCode:1, transactionId: 3, shoutId: 2
        case 3:
          println(shout.id +" "+ shout.remainingQuantity + " " + shout.limitPrice + " "+ shout.modReasonCode)
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9.00, shout.limitPrice)
          assertEquals(ModReasonCode.DELETIONBYUSER, shout.modReasonCode)
          assertNotSame(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }


   //* 2.) 2 TransactionLog entries:
    assertEquals(3, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: null, AskSize: 0, bid: 8.00, bidSize: 300, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: 9.00, AskSize: 150, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(9.00, translog.bid)
          assertEquals(150, translog.bidSize)
          assertNotSame(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: 8.00, AskSize: 3000, transactionId: 1
        case 3:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertNotSame(buy1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 3 Orderbook entries:
   assertEquals(3, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 8, bidSize1: 300, all other bid* and ask*==null, all other bidSize* and askSize*=0, transactionId: 1
        case 1:
          println(orderbook.bid1 + " "+ orderbook.bidSize1 + " "+ orderbook.ask1 + " "+ orderbook.askSize1)
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          break

        //bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 2:
          println(orderbook.bid1 + " "+ orderbook.bidSize1 + " "+orderbook.bid2 + " "+ orderbook.bidSize2 + " "+ orderbook.ask1 + " "+ orderbook.askSize1)
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          break

       //bid1: 8 EUR, bidSize1: 300, bid2: null, bidSize2: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 3:
          println(orderbook.bid1 + " "+ orderbook.bidSize1 + " "+ orderbook.ask1 + " "+ orderbook.askSize1)
          assertNotSame(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
    }
  }

  /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR
   * Action: Delete Order 1.)
   */

  void testDelFirstOfTwoBuyOrders() {
   //init
    buy1.remainingQuantity = 300
    buy1.limitPrice = 8.00
    auctionService.addShout(buy1)

    buy2.remainingQuantity = 150
    buy2.limitPrice = 9.00
    auctionService.addShout(buy2)

    //action
    auctionService.deleteShout(buy1)

    //* 1.) 3 shout objects in db
    assertEquals(3, Shout.list().size())

    def i=0
    for (Shout shout : Shout.listOrderById()){
      switch (++i){
        //Buy 300@8EUR, ModeReasonCode:1, transactionId: 1, shoutId: 1
        case 1:
          println(shout.id +" "+ shout.remainingQuantity + " " + shout.limitPrice + " "+ shout.modReasonCode)
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        //Buy 150@9EUR, ModeReasonCode:1, transactionId: 2, shoutId: 2
        case 2:
          println(shout.id +" "+ shout.remainingQuantity + " " + shout.limitPrice + " "+ shout.modReasonCode)
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(150, shout.remainingQuantity)
          assertEquals(9.00, shout.limitPrice)
          assertEquals(ModReasonCode.INSERT, shout.modReasonCode)
          assertEquals(buy2.transactionID, shout.transactionID)
          assertEquals(buy2.shoutID, shout.shoutID)
          break

        //DEL 150@9EUR, ModeReasonCode:1, transactionId: 3, shoutId: 2
        case 3:
          println(shout.id +" "+ shout.remainingQuantity + " " + shout.limitPrice + " "+ shout.modReasonCode)
          assertEquals(BuySellIndicator.BUY, shout.buySellIndicator)
          assertEquals(300, shout.remainingQuantity)
          assertEquals(8.00, shout.limitPrice)
          assertEquals(ModReasonCode.DELETIONBYUSER, shout.modReasonCode)
          assertNotSame(buy1.transactionID, shout.transactionID)
          assertEquals(buy1.shoutID, shout.shoutID)
          break

        default:
          assert("Unexpected shout")
      }
    }


   //* 2.) 2 TransactionLog entries:
    assertEquals(2, TransactionLog.list().size())

    i=0
    for (TransactionLog translog : TransactionLog.listOrderById()){
      switch (++i){
        //Quote: Ask: null, AskSize: 0, bid: 8.00, bidSize: 300, transactionId: 1
        case 1:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(8, translog.bid)
          assertEquals(300, translog.bidSize)
          assertEquals(buy1.transactionID, translog.transactionID)
          break

        //Quote: Bid: null, BidSize: 0, Ask: 9.00, AskSize: 150, transactionId: 2
        case 2:
          assertEquals(TransactionType.QUOTE, translog.type)
          assertNull(translog.ask)
          assertEquals(0, translog.askSize)
          assertEquals(9.00, translog.bid)
          assertEquals(150, translog.bidSize)
          assertNotSame(buy1.transactionID, translog.transactionID)
          break

        default:
          assert("Unexpected transactionlog")
      }
    }

   // 3.) 3 Orderbook entries:
   assertEquals(3, Orderbook.list().size())

   i=0
   for (Orderbook orderbook : Orderbook.listOrderById()){
     switch (++i){
        //bid1: 8, bidSize1: 300, all other bid* and ask*==null, all other bidSize* and askSize*=0, transactionId: 1
        case 1:
          println(orderbook.bid1 + " "+ orderbook.bidSize1 + " "+ orderbook.ask1 + " "+ orderbook.askSize1)
          assertEquals(buy1.transactionID, orderbook.transactionID)
          assertEquals(8, orderbook.bid1)
          assertEquals(300, orderbook.bidSize1)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          break

        //bid1: 9 EUR, bidSize1: 150, bid2: 8 EUR, bidSize2: 300, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 2:
          println(orderbook.bid1 + " "+ orderbook.bidSize1 + " "+orderbook.bid2 + " "+ orderbook.bidSize2 + " "+ orderbook.ask1 + " "+ orderbook.askSize1)
          assertEquals(buy2.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertEquals(8, orderbook.bid2)
          assertEquals(300, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          assertNull(orderbook.bid3)
          assertEquals(0, orderbook.bidSize3)
          break

       //bid1: 9 EUR, bidSize1: 300, bid2: null, bidSize2: 0, ask1: null, askSize1: 0, all other ask* and bid*==null, all other bidSize* and askSize*=0, transactionId: 2
        case 3:
          println(orderbook.bid1 + " "+ orderbook.bidSize1 + " "+ orderbook.ask1 + " "+ orderbook.askSize1)
          assertNotSame(buy1.transactionID, orderbook.transactionID)
          assertEquals(9, orderbook.bid1)
          assertEquals(150, orderbook.bidSize1)
          assertNull(orderbook.bid2)
          assertEquals(0, orderbook.bidSize2)
          assertNull(orderbook.ask1)
          assertEquals(0, orderbook.askSize1)
          break

        default:
          assert("Unexpected orderbook")
     }
    }
  }

   /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR
   * Action: Mod 1.) to Buy 300@9EUR
   */

   /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR
   * Action: Mod 1.) to Buy 300@7EUR
   */


   /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Buy 150@9EUR
   * Action: Mod 1.) to Buy 300@10EUR
   */

   /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Sell 150@9EUR
   * Action: Mod 1.) to Buy 300@9EUR
   */

   /**
   * Init: Create and save (in order):  1.) Buy 300@8EUR 2.) Sell 150@9EUR
   * Action: Mod 2.) to Sell 150@7EUR
   */

   void testSettelementOneTradeFullExecution() {
   //init
    sell1.limitPrice = 10.0
    sell1.remainingQuantity = 10.0
    auctionService.addShout(sell1)
    //action
    buy1.limitPrice = 10.0
    buy1.remainingQuantity = 10.0
    auctionService.addShout(buy1)

    assertEquals (2, DepotPosition.list().size())
    assertEquals (2, CashPosition.list().size())

    def cpBuy = CashPosition.findByPersonAndLatest(sampleBuyer, true)
    def cpSell = CashPosition.findByPersonAndLatest(sampleSeller, true)

    assertEquals(100, cpSell.getChange())
    assertEquals(100, cpSell.getBalance())

    assertEquals(-100, cpBuy.getChange())
    assertEquals(-100, cpBuy.getBalance())

     DepotPosition dpBuy = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleBuyer)
      eq('product', sampleProduct)
      eq('latest', true)
    }

     assertEquals(10, dpBuy.getChange())
     assertEquals(10, dpBuy.getBalance())

    def dpSell = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleSeller)
      eq('product', sampleProduct)
      eq('latest', true)
    }

    assertEquals(-10, dpSell.getChange())
    assertEquals(-10, dpSell.getBalance())

   }

  void testSettelementTwoTradesFullExecution() {
    //init
    sell1.limitPrice = 10.0
    sell1.remainingQuantity = 10.0
    auctionService.addShout(sell1)

    buy1.limitPrice = 10.0
    buy1.remainingQuantity = 10.0
    auctionService.addShout(buy1)

    sell2.limitPrice = 5.0
    sell2.remainingQuantity = 5.0
    auctionService.addShout(sell2)

    buy2.limitPrice = 5.0
    buy2.remainingQuantity = 5.0
    auctionService.addShout(buy2)

    assertEquals(4, DepotPosition.list().size())
    assertEquals(4, CashPosition.list().size())

    def cpBuy1 = CashPosition.findByPersonAndLatest(sampleBuyer, false)
    def cpBuy2 = CashPosition.findByPersonAndLatest(sampleBuyer, true)

    def cpSell1 = CashPosition.findByPersonAndLatest(sampleSeller, false)
    def cpSell2 = CashPosition.findByPersonAndLatest(sampleSeller, true)

    assertEquals(100, cpSell1.getChange())
    assertEquals(100, cpSell1.getBalance())

    assertEquals(-100, cpBuy1.getChange())
    assertEquals(-100, cpBuy1.getBalance())

    assertEquals(25, cpSell2.getChange())
    assertEquals(125, cpSell2.getBalance())

    assertEquals(-25, cpBuy2.getChange())
    assertEquals(-125, cpBuy2.getBalance())

    DepotPosition dpBuy1 = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleBuyer)
      eq('product', sampleProduct)
      eq('latest', false)
    }

    DepotPosition dpBuy2 = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleBuyer)
      eq('product', sampleProduct)
      eq('latest', true)
    }

    assertEquals(10, dpBuy1.getChange())
    assertEquals(10, dpBuy1.getBalance())

    assertEquals(5, dpBuy2.getChange())
    assertEquals(15, dpBuy2.getBalance())

    def dpSell1 = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleSeller)
      eq('product', sampleProduct)
      eq('latest', false)
    }

    def dpSell2 = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleSeller)
      eq('product', sampleProduct)
      eq('latest', true)
    }
    assertEquals(-10, dpSell1.getChange())
    assertEquals(-10, dpSell1.getBalance())

    assertEquals(-5, dpSell2.getChange())
    assertEquals(-15, dpSell2.getBalance())

  }

   void testSettlementTradePartialExecution() {
     //init
     sell1.limitPrice = 10.0
     sell1.remainingQuantity = 10.0
     auctionService.addShout(sell1)
     //action
     buy1.limitPrice = 10.0
     buy1.remainingQuantity = 5.0
     auctionService.addShout(buy1)

     assertEquals(2, DepotPosition.list().size())
     assertEquals(2, CashPosition.list().size())

     def cpBuy = CashPosition.findByPersonAndLatest(sampleBuyer, true)
     def cpSell = CashPosition.findByPersonAndLatest(sampleSeller, true)

     assertEquals(50, cpSell.getChange())
     assertEquals(50, cpSell.getBalance())

     assertEquals(-50, cpBuy.getChange())
     assertEquals(-50, cpBuy.getBalance())

     DepotPosition dpBuy = DepotPosition.withCriteria(uniqueResult: true) {
       eq('person', sampleBuyer)
       eq('product', sampleProduct)
       eq('latest', true)
     }

     assertEquals(5, dpBuy.getChange())
     assertEquals(5, dpBuy.getBalance())

     def dpSell = DepotPosition.withCriteria(uniqueResult: true) {
       eq('person', sampleSeller)
       eq('product', sampleProduct)
       eq('latest', true)
     }

     assertEquals(-5, dpSell.getChange())
     assertEquals(-5, dpSell.getBalance())

   }

  void testSettlementTradeSamePerson() {
    //init
    sell1.limitPrice = 10.0
    sell1.remainingQuantity = 10.0
    auctionService.addShout(sell1)
    //action
    buy1.limitPrice = 10.0
    buy1.remainingQuantity = 10.0
    buy1.setPerson(sampleSeller)
    auctionService.addShout(buy1)

    assertEquals(2, DepotPosition.list().size())
    assertEquals(2, CashPosition.list().size())

    def cp = CashPosition.findByPersonAndLatest(sampleSeller, true)
    assertEquals(0, cp.getBalance())


    def dp = DepotPosition.withCriteria(uniqueResult: true) {
      eq('person', sampleSeller)
      eq('product', sampleProduct)
      eq('latest', true)
    }

    assertEquals(0, dp.getBalance())

  }
}